import json

from socialgene.utils.logging import log


class SerializeToClustermap:
    """
    Take a sg_object and serialize all BGC object to clustermap.js format
    """

    def __init__(self, sg_object, sorted_bgcs, link_df, group_df):
        self._sg_object = sg_object
        self.sorted_bgcs = sorted_bgcs
        self._link_df = link_df
        self._group_df = group_df
        self._reset()

    def _reset(self):
        self._uid = 0
        self._uid_dict = {}
        self.feature_to_cmap_uid_dict = {}  # {"feature_obj": "clustermap_uid"}

    def _flatten_list(x):
        return sorted([item for row in x for item in row], reverse=True)

    def _get_uid(self, obj=None):
        # increment then return a uid as a string
        if obj in self._uid_dict:
            return str(self._uid_dict[obj])
        else:
            self._uid += 1
            self._uid_dict[str(self._uid)] = obj
            return str(self._uid)

    def _build(self):
        self._reset()
        return self._clusters() | self._create_links_dict() | self._create_groups_dict()

    def _clusters(self):
        log.info("Creating clustermap.js clusters")
        return {
            "clusters": [
                {
                    "uid": self._get_uid(obj=assembly),
                    "name": assembly.name,
                    "loci": self._loci([i for i in assembly.gene_clusters]),
                }
                for assembly in self.sorted_bgcs
            ]
        }

    def _feature(self, feature_obj):
        feat_id = self._get_uid(obj=feature_obj)
        self.feature_to_cmap_uid_dict.update({feature_obj: feat_id})
        return {
            "uid": feat_id,
            "label": feature_obj.external_id,
            "names": {
                "name": feature_obj.external_id,
                "description": feature_obj.description,
                "id": feature_obj.external_id,
                "locus_tag": feature_obj.locus_tag,
            },
            "start": feature_obj.start,
            "end": feature_obj.end,
            "strand": feature_obj.strand,
        }

    def _locus(self, locus_name, locus_obj):
        return {
            "uid": self._get_uid(obj=locus_obj),
            "name": locus_name,
            "genes": [self._feature(feature_obj) for feature_obj in locus_obj.features],
            "start": min((i.start for i in locus_obj.features)),
            "end": max((i.end for i in locus_obj.features)),
        }

    def _loci(self, loci):
        return [
            self._locus(locus_name=gc.parent.external_id, locus_obj=gc) for gc in loci
        ]

    def _create_groups_dict(self):
        """
        Create a dictionary of clustermap groups.

        Returns:
            A dictionary containing a list of groups, where each group is represented as a dictionary
            with keys "uid", "label", and "genes". e.g. {"groups": [{"uid": "1", "label": "group1", "genes": ["1", "2", "3"]}]
        """
        # Look in the clustermap uid dict (feature_to_cmap_uid_dict) for the cmap uid of each feature (query & target) in each group
        log.info("Creating clustermap.js links")
        groups = (
            self._group_df.groupby(["query_feature"])["target_feature"]
            .apply(list)
            .reset_index()
        )
        return {
            "groups": [
                {
                    "uid": self._get_uid(
                        obj=f"{i['query_feature'].external_id} {i['query_feature'].description}"
                    ),
                    "label": f"{i['query_feature'].external_id} {i['query_feature'].description}",
                    "genes": [self.feature_to_cmap_uid_dict[i["query_feature"]]]
                    + [
                        self.feature_to_cmap_uid_dict[i]
                        for i in i.target_feature
                        if i in self.feature_to_cmap_uid_dict
                    ],
                }
                for x, i in groups.iterrows()
            ]
        }

    def _create_links_dict(self):
        log.info("Creating clustermap.js links")

        return {
            "links": [
                {
                    "uid": self._get_uid(obj=None),
                    "target": {
                        "uid": self.feature_to_cmap_uid_dict[i["target_feature"]],
                        "name": self.feature_to_cmap_uid_dict[i["target_feature"]],
                    },
                    "query": {
                        "uid": self.feature_to_cmap_uid_dict[i["query_feature"]],
                        "name": self.feature_to_cmap_uid_dict[i["query_feature"]],
                    },
                    "identity": i.pident / 100 if "pident" in i else i.score,
                }
                for x, i in self._link_df.iterrows()
                if i["query_feature"] in self.feature_to_cmap_uid_dict
                and i["target_feature"] in self.feature_to_cmap_uid_dict
            ]
        }

    def write(self, outpath):
        log.info(f"Writing clustermap.js output to: {outpath}")
        with open(outpath, "w") as outfile:
            json.dump(
                self._build(),
                outfile,
                indent=4,
            )



import re

from socialgene.neo4j.neo4j_element import Node


class Publication(Node):
    neo4j_label = ["publication"]
    description = "Represents a publication"
    property_specification = {
        "doi": str,
        "pmid": int,
        "authors": str,
        "title": str,
        "journal": str,
        "year": int,
    }
    required_properties = ["doi"]
    constraints_unique = ["doi"]

    @staticmethod
    def _extract_doi(input):
        return re.search("10.\\d{4,9}/[-._;()/:a-z0-9A-Z]+", input).group()



import asyncio
import concurrent.futures
import re
from math import ceil
from pathlib import Path
from typing import List

import pandas as pd
from neo4j import AsyncGraphDatabase
from rich.progress import (
    BarColumn,
    MofNCompleteColumn,
    Progress,
    SpinnerColumn,
    TextColumn,
    TimeElapsedColumn,
)
from rich.table import Table

from socialgene.compare_proteins.hmmer import CompareDomains
from socialgene.config import env_vars
from socialgene.neo4j.neo4j import GraphDriver
from socialgene.search.base import SearchBase
from socialgene.utils.logging import CONSOLE, log

progress_bar = Progress(
    TextColumn("Ingesting target clusters from database..."),
    TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
    BarColumn(),
    MofNCompleteColumn(),
    TextColumn("• Time elapsed "),
    TimeElapsedColumn(),
)


def _find_similar_proteins(
    domain_list, frac: float = 0.75, only_culture_collection: bool = False
):
    """
    The function `_find_similar_proteins` is a synchronous function that queries a Neo4j graph database to
    find similar proteins based on protein domain information.

    Args:
      protein_domain_dict (Dict[List[str]]): The `protein_domain_dict` parameter is a dictionary where
    the keys are protein uids and the values are lists of domain uids.
      frac (float): The `frac` parameter is a float value that represents the fraction of protein
    domains that need to match in order for a protein to be considered similar. By default, it is set to
    0.75, meaning that at least 75% of the protein domains need to match

    Returns:
      The function `_find_sim_protein` returns a Pandas DataFrame containing the results of the query
    executed in the Neo4j database. The DataFrame has columns `assembly_uid`, `nucleotide_uid`,
    `target`, `n_start`, and `n_end`
    """
    # TODO: move async driver to reg driver class module
    # with GraphDriver() as driver:
    with GraphDriver().driver.session() as driver:
        res = driver.run(
            f"""
                WITH $domain_list AS input_protein_domains
                MATCH (prot1:protein)<-[a1:ANNOTATES]-(h0:hmm)
                WHERE h0.uid IN input_protein_domains
                WITH input_protein_domains, prot1, count(DISTINCT(h0)) as initial_count
                WHERE initial_count > size(input_protein_domains) * $frac
                MATCH (n1:nucleotide)-[e1:ENCODES]->(prot1)
                MATCH (a1:assembly)<-[:ASSEMBLES_TO]-(n1)
                {'WHERE (a1)-[:FOUND_IN]->(:culture_collection)' if only_culture_collection else ''}
                RETURN a1.uid as assembly_uid, n1.uid as nucleotide_uid, prot1.uid as target, e1.start as n_start, e1.end as n_end
                """,
            domain_list=list(domain_list),
            frac=frac,
        ).to_df()
        return res


async def _find_similar_proteins_async(
    domain_list, frac: float = 0.75, only_culture_collection: bool = False
):
    """
    The function `_find_similar_proteins_async` is an asynchronous function that queries a Neo4j graph database to
    find similar proteins based on protein domain information.

    Args:
      protein_domain_dict (Dict[List[str]]): The `protein_domain_dict` parameter is a dictionary where
    the keys are protein uids and the values are lists of domain uids.
      frac (float): The `frac` parameter is a float value that represents the fraction of protein
    domains that need to match in order for a protein to be considered similar. By default, it is set to
    0.75, meaning that at least 75% of the protein domains need to match

    Returns:
      The function `_find_sim_protein` returns a Pandas DataFrame containing the results of the query
    executed in the Neo4j database. The DataFrame has columns `assembly_uid`, `nucleotide_uid`,
    `target`, `n_start`, and `n_end`
    """
    # TODO: move async driver to reg driver class module
    async with AsyncGraphDatabase.driver(
        env_vars["NEO4J_URI"],
        auth=(
            env_vars["NEO4J_USER"],
            env_vars["NEO4J_PASSWORD"],
        ),
    ) as driver:
        res = await driver.execute_query(
            f"""
                WITH $domain_list AS input_protein_domains
                MATCH (prot1:protein)<-[a1:ANNOTATES]-(h0:hmm)
                WHERE h0.uid IN input_protein_domains
                WITH input_protein_domains, prot1, count(DISTINCT(h0)) as initial_count
                WHERE initial_count > size(input_protein_domains) * $frac
                MATCH (n1:nucleotide)-[e1:ENCODES]->(prot1)
                MATCH (a1:assembly)<-[:ASSEMBLES_TO]-(n1)
                {'WHERE (a1)-[:FOUND_IN]->(:culture_collection)' if only_culture_collection else ''}
                RETURN a1.uid as assembly_uid, n1.uid as nucleotide_uid, prot1.uid as target, e1.start as n_start, e1.end as n_end
                """,
            domain_list=list(domain_list),
            frac=frac,
        )
        res = pd.DataFrame(res.records, columns=res.keys)
        return res


sema = asyncio.BoundedSemaphore(5)


async def _find_similar_proteins_async_multiple(
    dict_of_domain_lists,
    frac: float = 0.75,
    only_culture_collection: bool = False,
):
    # create task group
    # TODO: if webserver in future this could be used to control max time of search
    async with sema:
        async with asyncio.TaskGroup() as group:
            # create and issue tasks
            tasks = {
                k: group.create_task(
                    _find_similar_proteins_async(
                        domain_list=v,
                        frac=frac,
                        only_culture_collection=only_culture_collection,
                    )
                )
                for k, v in dict_of_domain_lists.items()
            }
        # wait for all tasks to complete...
        # report all results
        # return tasks
        return pd.concat([v.result().assign(query=k) for k, v in tasks.items()])


def _find_similar_proteins_sync_multiple(
    dict_of_domain_lists,
    frac: float = 0.75,
    only_culture_collection: bool = False,
):
    results = []
    with concurrent.futures.ThreadPoolExecutor(
        max_workers=20,
    ) as executor:
        futures = {}
        for k, v in dict_of_domain_lists.items():
            future = executor.submit(
                _find_similar_proteins,
                domain_list=v,
                frac=frac,
                only_culture_collection=only_culture_collection,
            )
            futures[future] = k
        for f in concurrent.futures.as_completed(futures):
            result = f.result()
            result["query"] = futures[f]
            results.append(result)
    return pd.concat(results)


def run_search(
    dict_of_domain_lists, frac: float = 0.75, only_culture_collection: bool = False
):
    for k, v in dict_of_domain_lists.items():
        _find_similar_proteins(
            domain_list=v,
            frac=frac,
            only_culture_collection=only_culture_collection,
        )


def run_async_search(
    dict_of_domain_lists, frac: float = 0.75, only_culture_collection: bool = False
):
    return asyncio.run(
        _find_similar_proteins_async_multiple(
            dict_of_domain_lists=dict_of_domain_lists,
            frac=frac,
            only_culture_collection=only_culture_collection,
        )
    )


def run_sync_search(
    dict_of_domain_lists, frac: float = 0.75, only_culture_collection: bool = False
):
    return _find_similar_proteins_sync_multiple(
        dict_of_domain_lists=dict_of_domain_lists,
        frac=frac,
        only_culture_collection=only_culture_collection,
    )


class SearchDomains(SearchBase, CompareDomains):
    """
    Class search for similar BGCs in a SocialGene database, using domains
    Args:
        hmm_dir (str): Path to directory containing HMM profiles (used to annotate input BGC proteins if they aren't in the database).
        use_neo4j_precalc (bool): Try to pull domains from the database or annotate all input proteins with HMMER
        modscore_cutoff (float): Minimum score cutoff for a hit to be considered significant. (modified, combined Levenshtein + Jaccard)
        **kwargs: Additional keyword arguments to pass to the parent class.
    Raises:
        ValueError: If neither sg_object nor gbk_path is provided.
    """

    def __init__(
        self,
        hmm_dir: str = None,
        input=None,
        use_neo4j_precalc: bool = True,
        modscore_cutoff: float = 0.8,
        **kwargs,
    ) -> None:
        super().__init__(input=input, modscore_cutoff=modscore_cutoff, **kwargs)
        self.hmm_dir = hmm_dir
        self.outdegree_df = pd.DataFrame
        # input sg_object or gbk_path
        self._annotate(hmm_dir=self.hmm_dir, use_neo4j_precalc=use_neo4j_precalc)
        if not self._check_for_hmm_outdegree():
            self._set_hmm_outdegree()
        self._get_outdegree_per_hmm_per_protein()

    def search(self, run_async=True, **kwargs):
        dict_of_domain_lists = (
            self.outdegree_df.groupby("protein_uid", observed=True)["hmm_uid"]
            .apply(list)
            .to_dict()
        )
        log.info(
            "Searching database for proteins with similar domain content and all of the genomes those are found in"
        )
        with Progress(
            SpinnerColumn(spinner_name="aesthetic", speed=0.2),
        ) as progress:
            task = progress.add_task("Progress...", total=2)
            progress.update(task, advance=1)
            if run_async:
                self.raw_search_results_df = run_async_search(
                    dict_of_domain_lists, **kwargs
                )
                self.raw_search_results_df["query"] = self.raw_search_results_df[
                    "query"
                ].astype("category")
                progress.update(task, advance=1)
            else:
                self.raw_search_results_df = run_sync_search(
                    dict_of_domain_lists, **kwargs
                )
                self.raw_search_results_df["query"] = self.raw_search_results_df[
                    "query"
                ].astype("category")
                progress.update(task, advance=1)
        log.info(
            f"Initial search returned {len(self.raw_search_results_df):,} proteins, found in {self.raw_search_results_df.assembly_uid.nunique():,} genomes"
        )

    def _annotate(self, hmm_dir, use_neo4j_precalc: bool = True):
        log.info("Annotating input proteins domains")
        for x in Path(hmm_dir).iterdir():
            if re.search(r"\.hmm$|hmm\.gz$", str(x)):
                self.sg_object.annotate(
                    hmm_filepath=x,
                    use_neo4j_precalc=use_neo4j_precalc,
                )

    @staticmethod
    def _check_for_hmm_outdegree() -> bool:
        """
        Checks a single HMM model to see if it has an outdegree property
        """
        with GraphDriver() as db:
            res = db.run(
                """
               MATCH (h1:hmm)
               return h1.outdegree limit 1
             """
            ).peek()
        if res.value():
            return True
        else:
            return False

    @staticmethod
    def _set_hmm_outdegree() -> None:
        """
        Writes outdegree onto all hmm nodes (not just [:ANNOTATES])
        """
        with GraphDriver() as db:
            _ = db.run(
                """
                MATCH (h1: hmm)
                SET h1.outdegree = apoc.node.degree.out(h1, "ANNOTATES")
                """
            )

    def _get_outdegree_per_hmm_per_protein(self) -> pd.DataFrame:
        """
        Get the outdegree for each domain in each protein in an input sg_object
        Args:
            sg_object (SocialGene): SocialGene object
        Returns:
            DataFrame: pd.DataFrame with columns ["protein_uid", "hmm_uid", "outdegree"]
        """
        log.info("Finding the outdegree of all domains from all input proteins")
        # Get the non-redundant set of all domains from all proteins
        doms = set()
        for v in self.sg_object.proteins.values():
            for i in v.domain_vector:
                doms.add(i)
        if doms == set():
            raise ValueError("No domains in sg_object.proteins")
        doms = list(doms)
        # Retrieve the outdegree for the non-redundant set of domains
        with GraphDriver() as db:
            db_res = db.run(
                """
               WITH $doms as doms
               MATCH (h1:hmm)
               WHERE h1.uid in doms
               RETURN h1.uid as hmm_uid, h1.outdegree as outdegree
             """,
                doms=doms,
            ).to_df()
        db_res["hmm_uid"] = db_res["hmm_uid"].astype("category")
        # Create a DF of proteins and their domains
        temp = pd.DataFrame(
            data=(
                {"protein_uid": k, "hmm_uid": i}
                for k, v in self.sg_object.proteins.items()
                for i in v.domain_vector
            ),
            dtype="category",
        )
        # return a merged df with columns ["protein_uid", "hmm_uid", "outdegree"]
        self.outdegree_df = temp.merge(db_res, how="left")
        self.outdegree_df.drop_duplicates(inplace=True, ignore_index=True)

    def _filter_max_outdegree(self, max_outdegree: int = None):
        """Filter out proteins with a higher outdegree than max_outdegree

        Args:
            max_outdegree (int): HMM model annotations with an outdegree higher than this will be dropped
        """
        # The elif != Nones are to prevent an incorrect argument from just returning the the full DF, which would be unintended
        if isinstance(max_outdegree, int) or isinstance(max_outdegree, float):
            m_start = self.outdegree_df["outdegree"].sum()
            log.info(
                f"'max_outdegree' is set to {max_outdegree:,}, will remove any domains with a higher outdegree"
            )
            self.outdegree_df = self.outdegree_df[
                self.outdegree_df["outdegree"] <= max_outdegree
            ]
            log.info(
                f"'max_outdegree' reduced the total outdegree from {m_start:,} to {self.outdegree_df['outdegree'].sum():,}"
            )
        elif max_outdegree is not None:
            raise ValueError

    def _filter_max_domains_per_protein(self, max_domains_per_protein: int = None):
        """Filter out proteins with a higher outdegree than max_outdegree

        Args:
            max_domains_per_protein (int): HMM model annotations with an outdegree higher than this will be dropped
        """
        # The elif != Nones are to prevent an incorrect argument from just returning the the full DF, which would be unintended
        if isinstance(max_domains_per_protein, int) or isinstance(
            max_domains_per_protein, float
        ):
            m_start = self.outdegree_df["outdegree"].sum()
            log.info(
                f"'max_domains_per_protein' is set to {max_domains_per_protein:,}, will remove domains from proteins from highest to lowest outdegree"
            )
            self.outdegree_df = (
                self.outdegree_df.sort_values(
                    "outdegree", ascending=True, ignore_index=True
                )
                .groupby("protein_uid", observed=True)
                .head(max_domains_per_protein)
            )
            log.info(
                f"'max_domains_per_protein' reduced the total outdegree from {m_start:,} to {self.outdegree_df['outdegree'].sum():,}"
            )
        elif max_domains_per_protein is not None:
            raise ValueError

    def _filter_max_query_proteins(self, max_query_proteins: int = None):
        """Filter out proteins with a higher outdegree than max_outdegree

        Args:
            max_query_proteins (int): HMM model annotations with an outdegree higher than this will be dropped
        """
        # The elif != Nones are to prevent an incorrect argument from just returning the the full DF, which would be unintended
        if isinstance(max_query_proteins, int) or isinstance(max_query_proteins, float):
            m_start = self.outdegree_df["outdegree"].sum()
            log.info(
                f"'max_query_proteins' is set to {max_query_proteins:,}, will limit search to {max_query_proteins} of {len(self.outdegree_df['protein_uid'].unique())} input proteins"
            )
            proteins_to_keep = (
                self.outdegree_df.sort_values(
                    "outdegree", ascending=True, ignore_index=True
                )
                .groupby("protein_uid", observed=True)
                .sum("outdegree")
                .sort_values("outdegree", ascending=True, ignore_index=False)
                .head(max_query_proteins)
                .index.to_list()
            )
            self.outdegree_df = self.outdegree_df[
                self.outdegree_df["protein_uid"].isin(proteins_to_keep)
            ]
            log.info(
                f"'max_query_proteins' reduced the total outdegree from {m_start:,} to {self.outdegree_df['outdegree'].sum():,}"
            )
        elif max_query_proteins is not None:
            raise ValueError

    def _filter_scatter(self, max_query_proteins):
        """Choose a random subset of proteins to search that are spread across the length of the input BGC."""
        log.info("Choosing query proteins that span across the input BGC")
        temp = list(self.input_bgc.features_sorted_by_midpoint)
        if max_query_proteins > len(temp):
            max_query_proteins = len(temp)
        temp = [
            temp[int(ceil(i * len(temp) / max_query_proteins))].uid
            for i in range(max_query_proteins)
        ]
        self.outdegree_df = self.outdegree_df[
            self.outdegree_df["protein_uid"].isin(temp)
        ]
        log.info("Scattering the search to proteins to and across the input BGC")
        log.info(
            f"'max_query_proteins' is set to {max_query_proteins:,}, will limit search to {max_query_proteins} of {len(self.outdegree_df['protein_uid'].unique())} input proteins"
        )

    def prioritize_input_proteins(
        self,
        max_query_proteins: float = None,
        max_domains_per_protein: int = None,
        max_outdegree: int = None,
        scatter: bool = False,
        locus_tag_bypass_list: List[str] = None,
        protein_id_bypass_list: List[str] = None,
    ):
        """Rank input proteins by how many (:hmm)-[:ANNOTATES]->(:protein) relationships will have to be traversed

        Args:
            df (pd.DataFrame): pd.DataFrame({"external_id":[], "hmm_uid":[], "outdegree":[]})
            max_query_proteins (float): Max proteins to return. If >0 and <1, will return X% of input proteins
            max_domains_per_protein (int): Max domains to retain for each individual protein (highest outdegree dropped first)
            max_outdegree (int): HMM model annotations with an outdegree higher than this will be dropped
            scatter (bool, optional): Choose a random subset of proteins to search that are spread across the length of the input BGC. Defaults to False.
            locus_tag_bypass_list (List[str], optional): List of locus tags that will bypass filtering. This is the ID found in a GenBank file "/locus_tag=" field.  Defaults to None.
            protein_id_bypass_list (List[str], optional): Less preferred than `bypass`. List of external protein IDs that will bypass filtering. This is the ID found in a GenBank file "/protein_id=" field. Defaults to None.
        Returns:
            pd.DataFrame: pd.DataFrame({"external_id":[], "hmm_uid":[], "outdegree":[]})
        """
        log.info("Prioritizing input proteins by outdegree")
        loci_protein_ids = set()
        if locus_tag_bypass_list:
            # bypass using locus tags
            loci_protein_ids = {
                i.uid
                for i in self.input_bgc.features
                if i.locus_tag in list(locus_tag_bypass_list)
            }
        if protein_id_bypass_list:
            # bypass using an external protein ids
            loci_protein_ids.update(
                {
                    i.uid
                    for i in self.input_bgc.features
                    if i.external_id in list(protein_id_bypass_list)
                }
            )
        bypass_df = self.outdegree_df[
            self.outdegree_df["protein_uid"].isin(loci_protein_ids)
        ]
        bypass_df = (
            bypass_df.sort_values("outdegree", ascending=True, ignore_index=True)
            .groupby("protein_uid", observed=True)
            .head(max_domains_per_protein)
        )
        len_start = self.outdegree_df["outdegree"].sum()
        if self.outdegree_df["outdegree"].sum() == 0:
            raise ValueError(
                "None of the domains of the input proteins have connections in the database"
            )
        if (self.outdegree_df["outdegree"] == 0).any():
            prelen = len(self.outdegree_df)
            prelen_prot = len(self.outdegree_df["protein_uid"].unique())
            self.outdegree_df = self.outdegree_df[self.outdegree_df["outdegree"] > 0]
            log.info(
                f"Removed {prelen - len(self.outdegree_df)} domains from consideration (no connections in the DB), which reemoves {prelen_prot - len(self.outdegree_df['protein_uid'].unique())} proteins from consideration"
            )

        if max_query_proteins and scatter:
            self._filter_scatter(max_query_proteins)

        self._filter_max_outdegree(max_outdegree)
        self._filter_max_domains_per_protein(max_domains_per_protein)

        if max_query_proteins and not scatter:
            # Make sure max_query_proteins accounts for bypassed proteins
            if bypass_df.protein_uid.nunique() > 0:
                log.warning(bypass_df)
                max_query_proteins = (
                    max_query_proteins - bypass_df.protein_uid.nunique()
                )
                if max_query_proteins < 0:
                    max_query_proteins = 0
            self._filter_max_query_proteins(max_query_proteins)

        # Add back explicitly requested input proteins/loci

        self.outdegree_df = pd.merge(self.outdegree_df, bypass_df, how="outer")
        log.info(
            f"The total outdegree was {len_start:,}; now it's {self.outdegree_df['outdegree'].sum():,}"
        )
        self.n_searched_proteins = len(self.outdegree_df["protein_uid"].unique())

    @property
    def outdegree_table(self):
        table = Table(title="Outdegree of input protein domains")
        table.add_column("Protein", justify="left", style="cyan", no_wrap=True, ratio=1)
        table.add_column(
            "Locus/Descripton", justify="left", style="cyan", no_wrap=True, ratio=1
        )
        table.add_column(
            "Unique HMM models", justify="left", style="cyan", no_wrap=True, ratio=1
        )
        table.add_column("Mean", justify="left", style="cyan", no_wrap=True, ratio=1)
        table.add_column("Min", justify="left", style="cyan", no_wrap=True, ratio=1)
        table.add_column("25%", justify="left", style="cyan", no_wrap=True, ratio=1)
        table.add_column("50%", justify="left", style="cyan", no_wrap=True, ratio=1)
        table.add_column("75%", justify="left", style="cyan", no_wrap=True, ratio=1)
        table.add_column("Max", justify="left", style="cyan", no_wrap=True, ratio=1)
        table.add_column("Sum", justify="left", style="cyan", no_wrap=True, ratio=1)
        for i in self._outdegree_table_stats():
            table.add_row(*[str(i) for i in i])
        before_width = CONSOLE.width
        CONSOLE.width = 300
        CONSOLE.print(table)
        CONSOLE.width = before_width

    def _outdegree_table_stats(self):
        temp = (
            self.outdegree_df.groupby("protein_uid", observed=True)["outdegree"]
            .describe()
            .reset_index()
        )
        temp.drop(
            "std",
            axis=1,
            inplace=True,
        )
        summed = (
            self.outdegree_df.groupby("protein_uid", observed=True)["outdegree"]
            .sum()
            .reset_index(name="sum")
        )
        temp = pd.merge(temp, summed, on="protein_uid")
        temp = temp.sort_values("sum", ascending=True, ignore_index=True)
        temp[list(["count", "mean", "min", "25%", "50%", "75%", "max", "sum"])] = temp[
            list(["count", "mean", "min", "25%", "50%", "75%", "max", "sum"])
        ].astype(int)
        d = [
            {
                "protein_uid": i.uid,
                "desc": f"{i.external_id} | {i.locus_tag if i.locus_tag else 'no-locus-tag'} | {i.description}",
            }
            for i in self.input_bgc.features
        ]
        d = pd.DataFrame(d)
        temp = pd.merge(temp, d, on="protein_uid", how="left")
        temp = temp[
            [
                "protein_uid",
                "desc",
                "count",
                "mean",
                "min",
                "25%",
                "50%",
                "75%",
                "max",
                "sum",
            ]
        ]
        return temp.values



import argparse
import os
from pathlib import Path

from rich.progress import Progress, SpinnerColumn

from socialgene.parsers.hmmmodel import HMM_SOURCES, HmmModelHandler
from socialgene.utils.logging import log

parser = argparse.ArgumentParser(description="Parse NcbiAssembliessdsd taxonomy")
parser.add_argument(
    "--input_dir",
    type=Path,
    help="input directory filepath",
    required=True,
)
parser.add_argument(
    "--outdir",
    type=Path,
    help="output directory filepath",
    required=True,
)

parser.add_argument(
    "--glob",
    default="**/*.socialgene.hmm.gz",
    help="Glob for hmm files",
    type=str,
    required=False,
)


def run_nf_workflow(input_dir, outdir, input_glob="**/*.socialgene.hmm.gz"):
    # glob determined by filepath assigned in the nextflow workflow: https://github.com/socialgene/sgnf/blob/main/bin/hmmconvert_loop.sh
    hmms_object = HmmModelHandler()
    # expects all HMM databases to be saved in a folder named for
    # the database in which it was downloaded from (e.g. antismash, pfam, etc.)
    # so that the name of the source can be assigned
    dirpaths = [
        os.path.join(input_dir, i) for i in os.listdir(input_dir) if i in HMM_SOURCES
    ]
    dirpaths.sort()
    hmm_paths = []
    for dirpath in dirpaths:
        # hmm_paths is a list of dicts: [{filepath:"", base_dir:""}]
        hmm_paths.extend(
            [
                {"filepath": i, "base_dir": dirpath}
                for i in sorted(Path(dirpath).glob(input_glob))
            ]
        )
    # sort for reproducibility
    # read and parse all models
    with Progress(
        SpinnerColumn(spinner_name="runner"),
        *Progress.get_default_columns(),
        transient=True,
    ) as progress:
        task = progress.add_task("Progress...", total=len(hmm_paths))
        for i in hmm_paths:
            hmms_object.read(filepath=i.get("filepath"), base_dir=i.get("base_dir"))
            log.debug(f"Parsing {i.get('filepath')}\r")
            progress.update(task, advance=1)
    # Update info, make non-redundant, and write out
    hmms_object.hydrate_cull()
    hmms_object.remove_duplicate_and_old_pfam()
    hmms_object.remove_duplicate_hash()
    hmms_object.write(
        outdir=outdir,
        hash_as_name=True,
    )
    hmms_object.write_metadata_tsv(outdir=outdir, header=False)
    hmms_object.write_hmm_node_tsv(outdir=outdir)


def main():
    args = parser.parse_args()
    input_dir = args.input_dir
    outdir = args.outdir

    run_nf_workflow(
        input_dir=input_dir,
        outdir=outdir,
        input_glob=args.glob,
    )


if __name__ == "__main__":
    main()



from math import floor, log10


def find_exp(number) -> int:
    number = float(number)
    if number < 0:
        raise ValueError("Function only meant for positive numbers")
    if number == 0:
        return 0
    base10 = log10(abs(number))
    return int(floor(base10))



#!/usr/bin/env python
import importlib.resources as importlib_resources
import os

os_vars = dict(os.environ)
env_vars = {}


env_file = importlib_resources.files("socialgene") / "common_parameters.env"


def strtobool(val):
    val = val.lower()
    if val == "true":
        return True
    elif val == "false":
        return False
    else:
        raise ValueError(f"invalid truth value {val}")


internal_vars = {}
with open(env_file) as f:
    for line in f:
        if line.startswith("#") or line == "\n":
            continue
        k, v = line.strip().split("=", 1)
        internal_vars[k] = v


for k, v in internal_vars.items():
    if isinstance(v, str):
        try:
            internal_vars[k] = int(v)
        except ValueError:
            internal_vars[k] = v


for k, v in internal_vars.items():
    if isinstance(v, str):
        try:
            internal_vars[k] = float(v)
        except ValueError:
            internal_vars[k] = v

for k, v in internal_vars.items():
    if isinstance(v, str) and v.lower() in ("true", "false"):
        try:
            internal_vars[k] = strtobool(v)
        except ValueError:
            internal_vars[k] = v


for k, v in internal_vars.items():
    if v == "NONE":
        internal_vars[k] = None


for k, v in internal_vars.items():
    if k in os_vars:
        env_vars[k] = os_vars[k]
    else:
        env_vars[k] = v



from ftplib import FTP
from io import BytesIO

from socialgene.utils.logging import log


class NcbiFtp:
    ftp = FTP("ftp.ncbi.nlm.nih.gov", "anonymous", "cmclark8@wisc.edu", timeout=60)

    def __init__(self):
        log.debug(
            f'Connected to "{self.ftp.host}" on {self.ftp.port};\n\n NCBI FTP info: {self.ftp.welcome.replace("220", "")}'
        )
        self.ftp.set_pasv(True)
        self.assembly_paths = []

    def _connect_to_ftp(self):
        log.debug("Connecting to NCBI FTP")
        self.ftp.login()

    def get_assembly_dirpaths(self):
        with BytesIO() as r:
            log.debug("Beginning download of assembly_summary_refseq.txt")
            self.ftp.retrbinary(
                "RETR /genomes/ASSEMBLY_REPORTS/assembly_summary_refseq.txt", r.write
            )
            self.assembly_paths = []
            for i in r.getvalue().decode("utf-8").split("\n"):
                if i.startswith("#"):
                    pass
                elif i == "":
                    pass
                else:
                    self.assembly_paths.append(i.split("\t")[19])
        self.assembly_paths = [
            i.replace("ftp://ftp.ncbi.nlm.nih.gov", "") for i in self.assembly_paths
        ]
        self.assembly_paths = [
            i.replace("https://ftp.ncbi.nlm.nih.gov", "") for i in self.assembly_paths
        ]
        self.assembly_paths = [i.strip("/") for i in self.assembly_paths]
        log.debug("Finished download of assembly_summary_refseq.txt")
        log.debug(f"Found {len(self.assembly_paths)} assembly urls")



"""https://np-mrd.org/"""

from socialgene.neo4j.neo4j_element import Node


class Npmrd(Node):
    neo4j_label = ["npmrd"]
    description = "Represents a single NP-MRD entry"
    property_specification = {
        "uid": str,
    }
    required_properties = ["uid"]
    constraints_unique = ["uid"]



import socialgene.utils.file_handling as fh
from socialgene.parsers.fasta import FastaParserMixin
from socialgene.parsers.genbank import GenbankParserMixin
from socialgene.parsers.gff import GFFParserMixin


class SequenceParser(GenbankParserMixin, FastaParserMixin, GFFParserMixin):
    def __init__(self):
        super().__init__()

    def parse(self, filepath, **kwargs):
        """Parse sequence files (main function)

        Args:
            filepath (str): path to sequence file
        """
        # TODO: fh.check_if_tar(filepath=filepath)
        if isinstance(filepath, str) and filepath.startswith(">"):
            self.parse_fasta_file(input=filepath, **kwargs)
            return
        else:
            filetype = fh.guess_filetype(filepath)
            if filetype == "genbank":
                self._open_genbank(input_path=filepath, **kwargs)
                return
            elif filetype == "fasta":
                self.parse_fasta_file(input=filepath, **kwargs)
                return
            elif filetype == "gff":
                self.parse_gff_file(input_path=filepath, **kwargs)
                return

        raise NotImplementedError(
            "May not be implemented, or you need to use the genbank/fasta parser directly. (e.g. for tar archives)"
        )



from abc import ABC, abstractmethod
from typing import List

import pandas as pd

from socialgene.base.molbio import Protein

BlastTab_COLUMNS = {
    "query": str,
    "target": str,
    "pident": "Float64",
    "length": "Int64",
    "mismatch": "Int64",
    "gapopen": "Int64",
    "qstart": "Int64",
    "qend": "Int64",
    "sstart": "Int64",
    "send": "Int64",
    "evalue": "Float64",
    "score": "Float64",
    "qlen": "Int64",
    "slen": "Int64",
}


class ProteinComparison(ABC):
    @abstractmethod
    def compare_proteins(self, p1: List[Protein], p2: List[Protein]): ...  # noqa: E704

    def reciprocal_hits(
        self,
        q_vs_t_df: pd.DataFrame,
        t_vs_q_df: pd.DataFrame,
    ):
        """This function takes a dataframe and returns a dataframe with the reciprocal best hits"""
        # Filter the DF, keep the best target hit for each query protein
        q_vs_t_df = self.best_hit_to_query(q_vs_t_df)
        # Filter the DF, keep the best query hit for each target protein
        t_vs_q_df = self.best_hit_to_query(t_vs_q_df)
        t_vs_q_df.rename(columns={"query": "target", "target": "query"}, inplace=True)
        df = pd.concat([q_vs_t_df, t_vs_q_df])
        del q_vs_t_df, t_vs_q_df
        df.sort_values(by=self.score_column, ascending=False, inplace=True)
        # remove non reciprocal
        df = df[df.duplicated(["query", "target"], keep=False)]
        df = df.drop_duplicates(subset=["query", "target", self.score_column])
        # keep the reciprocal hit with the highest bitscore
        # but if multiple hits have the same highest bitscore, keep them all
        return df[
            df.groupby(["query", "target"])[self.score_column].transform("max")
            == df[self.score_column]
        ]

    def best_hit_to_query(
        self,
        df: pd.DataFrame,
    ):
        return (
            df.reset_index(inplace=False, drop=True)
            .sort_values(self.score_column, ascending=False)
            .drop_duplicates("query", keep="first")
            .reset_index(inplace=False, drop=True)
        )


class HMMDataFrame(ProteinComparison):
    def __init__(self, score_column="score"):
        self.score_column = score_column


class BlastTab(ProteinComparison):
    def __init__(self, score_column="score"):
        self.score_column = score_column
        self.score_scale = float("inf")



"""https://mibig.secondarymetabolites.org/"""

import re

from socialgene.addons.chemistry.nr import ChemicalCompoundNode
from socialgene.neo4j.neo4j_element import Node, Relationship
from socialgene.nextflow.nodes import ASSEMBLY


class Mibig(ASSEMBLY):
    neo4j_label = ["assembly", "mibig_bgc"]
    description = "Represents a single Mibig entry"
    property_specification = {
        "uid": str,
    }
    required_properties = ["uid"]

    def __init__(self, properties=None) -> None:
        super().__init__()
        if properties:
            self.properties["uid"] = self.attempt_to_fix_uid(properties["uid"])

    @staticmethod
    def attempt_to_fix_uid(uid):
        # npatlas has some mibig ids prefixed, some not, try to add prefix here
        # https://github.com/NPAtlas/npatlas_website/issues/13
        if uid.startswith("BGC"):
            temp = uid
        else:
            temp = "BGC" + uid.zfill(7)
        # Validate uid
        if re.match("BGC[0-9]{7}", temp):
            return temp
        else:
            raise ValueError(f"Unexpected mibig ID {uid}")


class Substrate(ChemicalCompoundNode):
    neo4j_label = ChemicalCompoundNode.neo4j_label + ["substrate"]
    description = "Mibig substrate (e.g. NRPS monomer)"


class MibigCompound(Node):
    neo4j_label = ["mibig_compound"]
    description = "Represents a single Mibig compound"
    property_specification = {
        "name": str,
        "smiles": str,
        "inchi": str,
    }
    required_properties = ["uid"]
    constraints_unique = ["uid"]


class MibigActivity(Node):
    neo4j_label = ["mibig_activity"]
    description = "Represents a single Mibig bioactivity"
    property_specification = {
        "uid": str,
    }
    required_properties = ["uid"]
    constraints_unique = ["uid"]


class Mibig_Biosynthetic_Class(Node):
    neo4j_label = ["mibig_biosynthetic_class"]
    description = "Represents a single Mibig biosynthetic class"
    property_specification = {
        "uid": str,
    }
    required_properties = ["uid"]
    constraints_unique = ["uid"]


class MibigToMibigCompound(Relationship):
    neo4j_label = "PRODUCES"
    description = "Connects a Mibig BGC to a Mibig compound"
    start_class = Mibig
    end_class = MibigCompound



from typing import List

from socialgene.neo4j.neo4j import Neo4jQuery

neo4j_object = Neo4jQuery()


def search_protein_hash(query_hash: List[str]):
    """Search the database for proteins with exact hash matches

    Args:
        query_hash (list): list of hash strings

    Raises:
        ValueError: incorrect input

    Returns:
        dict: dictionary of whether the protein hash is in the database, e.g. {query1hash: True, query2hash: False}
    """
    if isinstance(query_hash, str):
        query_hash = [query_hash]
    elif not isinstance(query_hash, list) and all(
        [isinstance(i, str) for i in query_hash]
    ):
        raise ValueError("query_hash must be a list of strings")
    # results will be e.g. [{'result': ['XcXHqKUXe3T2zDq3bhJxuw_0rJ2zwreU']}]
    results = neo4j_object.query_neo4j(
        cypher_name="search_protein_hash", param=query_hash
    )
    # Create a dictionary of all input values, with False as the default
    query_hash_dict = {i: False for i in query_hash}
    # Change returned ids to True
    for i in results[0]["result"]:
        query_hash_dict[i] = True
    return query_hash_dict



import argparse

from socialgene.dbmodifiers.massage.massage import (
    add_antismash_regions_as_edges,
    add_antismash_regions_as_nodes,
    add_protein_descriptions,
    add_taxonomic_name_to_assembly,
    culture_collections_as_nodes_rels,
    delete_antismash_regions_as_nodes,
    fix_mibig_taxonomy,
    set_mibig_bgc,
)
from socialgene.utils.logging import log

NCBI_RANKS = [
    "biotype",
    "clade",
    "class",
    "cohort",
    "family",
    "forma",
    "forma specialis",
    "genotype",
    "genus",
    "infraclass",
    "infraorder",
    "isolate",
    "kingdom",
    "morph",
    "no rank",
    "order",
    "parvorder",
    "pathogroup",
    "phylum",
    "section",
    "series",
    "serogroup",
    "serotype",
    "species",
    "species group",
    "species subgroup",
    "strain",
    "subclass",
    "subcohort",
    "subfamily",
    "subgenus",
    "subkingdom",
    "suborder",
    "subphylum",
    "subsection",
    "subspecies",
    "subtribe",
    "superclass",
    "superfamily",
    "superkingdom",
    "superorder",
    "superphylum",
    "tribe",
    "varietas",
]
parser = argparse.ArgumentParser(description="Parse NcbiAssembliessdsd taxonomy")

parser.add_argument(
    "--antismash_as_edges",
    help="",
    default=False,
    required=False,
    action=argparse.BooleanOptionalAction,
)
parser.add_argument(
    "--antismash_as_nodes",
    help="",
    default=False,
    required=False,
    action=argparse.BooleanOptionalAction,
)
parser.add_argument(
    "--delete_antismash_regions_as_nodes",
    help="",
    default=False,
    required=False,
    action=argparse.BooleanOptionalAction,
)
parser.add_argument(
    "--protein_descriptions",
    help="",
    default=False,
    required=False,
    action=argparse.BooleanOptionalAction,
)
parser.add_argument(
    "--fix_mibig_taxonomy",
    help="",
    default=False,
    required=False,
    action=argparse.BooleanOptionalAction,
)
parser.add_argument(
    "--set_mibig_bgc",
    help="",
    default=False,
    required=False,
    action=argparse.BooleanOptionalAction,
)
parser.add_argument(
    "--culture_collections",
    help="",
    default=False,
    required=False,
    action=argparse.BooleanOptionalAction,
)

parser.add_argument(
    "--taxa",
    nargs="+",
    type=str,
    help="Adds taxon as a node property of the given rank for each assembly (eg: genus family)",
    required=False,
    default=None,
    choices=NCBI_RANKS,
)


def main():
    args = parser.parse_args()
    # defaults are all negative, print help if nothing is provided
    if not any([i for i in args.__dict__.values()]):
        parser.print_help()
        return
    if args.antismash_as_edges:
        add_antismash_regions_as_edges()
    if args.antismash_as_nodes:
        add_antismash_regions_as_nodes()
    if args.delete_antismash_regions_as_nodes:
        delete_antismash_regions_as_nodes()
    if args.protein_descriptions:
        add_protein_descriptions()
    if args.fix_mibig_taxonomy:
        fix_mibig_taxonomy()
    if args.set_mibig_bgc:
        set_mibig_bgc()
    if args.culture_collections:
        culture_collections_as_nodes_rels()
    if args.taxa:
        for i in args.taxa:
            if i in NCBI_RANKS:
                add_taxonomic_name_to_assembly(i)
            else:
                log.warning(f"{i} not in the list of known ranks")


if __name__ == "__main__":
    main()



"""http://classyfire.wishartlab.com"""

from socialgene.addons.chebi.nr import ChebiNode
from socialgene.neo4j.neo4j_element import Node, Relationship


class ClassyFireNode(Node):
    neo4j_label = ["classyfire"]
    description = "Represents a classyfire chemical ontology term"
    property_specification = {"uid": int, "name": str, "definition": str}
    required_properties = ["uid"]
    constraints_unique = ["uid"]


class ClassyFireIsA(Relationship):
    neo4j_label = "IS_A"
    description = "Represents a relationship between chemont nodes"
    start_class = ClassyFireNode
    end_class = ClassyFireNode


class ClassyFireSynonym(Relationship):
    neo4j_label = "SYNONYM"
    description = "Represents a synonym relationship between chemont chebi"
    start_class = ClassyFireNode
    end_class = ChebiNode



from socialgene.neo4j.neo4j import GraphDriver
from socialgene.utils.logging import log


def _add_to_neo4j(statement, **kwargs):
    try:
        (
            GraphDriver().driver.execute_query(
                statement,
                **kwargs,
                database_="neo4j",
            )
        )
    except Exception as e:
        # If fails because index exists, don't fail, just report it. Fail any other errors.
        if e.code == "Neo.ClientError.Schema.EquivalentSchemaRuleAlreadyExists":
            log.warning(e.message)
        else:
            raise e


def mess(func):
    def wrapper():
        log.info("May appear to hang while index/constraint is populating")
        func()

    return wrapper


def constraint_status():
    with GraphDriver() as db:
        print(db.run("SHOW CONSTRAINTS").to_df())


def index_status():
    with GraphDriver() as db:
        print(db.run("SHOW INDEXES").to_df())


@mess
def assembly_uid():
    _add_to_neo4j(
        """
            CREATE CONSTRAINT assembly_uid IF NOT EXISTS
            FOR (n:assembly)
            REQUIRE (n.uid) IS UNIQUE;
            """
    )


@mess
def protein_uid():
    _add_to_neo4j(
        """
            CREATE CONSTRAINT protein_uid IF NOT EXISTS
            FOR (n:protein)
            REQUIRE (n.uid) IS UNIQUE;
            """
    )


@mess
def hmm_uid():
    _add_to_neo4j(
        """
            CREATE CONSTRAINT hmm_uid IF NOT EXISTS
            FOR (n:hmm)
            REQUIRE (n.uid) IS UNIQUE;
            """
    )


@mess
def nucleotide_uid():
    _add_to_neo4j(
        """
            CREATE CONSTRAINT nucleotide_uid IF NOT EXISTS
            FOR (n:nucleotide)
            REQUIRE (n.uid) IS UNIQUE;
            """
    )


@mess
def nucleotide_external_id():
    _add_to_neo4j(
        """
            CREATE INDEX nuc_extern_id
            FOR (n:nucleotide)
            ON (n.external_id);
            """
    )


@mess
def taxonomy_uid():
    _add_to_neo4j(
        """
            CREATE INDEX taxonomy_uid
            FOR (n:taxid)
            ON (n.uid);
            """
    )


@mess
def gnps_library_uid():
    _add_to_neo4j(
        """
            CREATE INDEX gnps_library_uid
            FOR (n:gnps_library_spectrum)
            ON (n.uid);
            """
    )


@mess
def npatlas_uid():
    _add_to_neo4j(
        """
            CREATE INDEX npatlas_uid
            FOR (n:npatlas)
            ON (n.uid);
            """
    )



from socialgene.neo4j.neo4j_element import Node, Relationship


class Product(Node):
    neo4j_label = ["product"]
    description = "Represents the product of an antiSMASH BGC"
    property_specification = {
        "uid": int,
    }
    constraints_unique = ["uid"]


class Category(Node):
    neo4j_label = ["category"]
    description = "Represents the category of an antiSMASH BGC"
    property_specification = {
        "uid": int,
    }
    constraints_unique = ["uid"]


class ProductToCategory(Relationship):
    neo4j_label = "IS_A"
    description = "Connects an antiSMASH product to category "
    start_class = Product
    end_class = Category


class GeneClusterToProduct(Relationship):
    neo4j_label = "IS_A"
    description = "Connects an antiSMASH BGC to product "
    start_class = Product
    end_class = Category
    property_specification = {
        "tool": str,
        "start": int,
        "end": int,
        "core_start": int,
        "core_end": int,
    }



import socialgene.nextflow.nodes as Nodes
from socialgene.neo4j.neo4j_element import Relationship

# This file can only contain class objects that inherit from Relationship


class ANNOTATES(Relationship):
    """Represents the relationship between a protein and a HMM"""

    neo4j_label = "ANNOTATES"
    description = "Represents the relationship between a protein and a HMM"
    header_filename = "protein_to_hmm_header.header"
    target_subdirectory = "parsed_domtblout"
    target_extension = "parseddomtblout"
    header = [
        ":END_ID(protein)",
        ":START_ID(hmm)",
        "env_from:Long",
        "env_to:Long",
        "seq_pro_score:Float",
        "evalue:Long",
        "i_evalue:Long",
        "domain_bias:Float",
        "domain_score:Float",
        "seq_pro_bias:Float",
        "hmm_from:Long",
        "hmm_to:Long",
        "ali_from:Long",
        "ali_to:Long",
        "exponentialized:Boolean",
    ]
    property_specification = {
        "env_from": float,
        "env_to": float,
        "seq_pro_score": float,
        "evalue": float,
        "i_evalue": float,
        "domain_bias": float,
        "domain_score": float,
        "seq_pro_bias": float,
        "hmm_from": float,
        "hmm_to": float,
        "ali_from": float,
        "ali_to": float,
        "exponentialized": bool,
    }
    start_class = Nodes.HMM
    end_class = Nodes.PROTEIN


class ASSEMBLES_TO(Relationship):
    """Represents the relationship between an assembly and a nucleotide sequence (e.g. scaffold, contig, chromosome)"""

    neo4j_label = "ASSEMBLES_TO"
    description = "Represents the relationship between an assembly and a nucleotide sequence (e.g. scaffold, contig, chromosome)"
    header_filename = "assembly_to_locus.header"
    target_subdirectory = "genomic_info"
    target_extension = "assembly_to_locus"
    header = [":END_ID(assembly)", ":START_ID(nucleotide)"]
    start_class = Nodes.NUCLEOTIDE
    end_class = Nodes.ASSEMBLY


class ENCODES(Relationship):
    """Represents the relationship between a nucleotide sequence and a protein"""

    neo4j_label = "ENCODES"
    description = (
        "Represents the relationship between a nucleotide sequence and a protein"
    )
    header_filename = "locus_to_protein.header"
    target_subdirectory = "genomic_info"
    target_extension = "locus_to_protein"
    header = [
        ":START_ID(nucleotide)",
        ":END_ID(protein)",
        "external_id",
        "locus_tag",
        "start:Long",
        "end:Long",
        "strand:Long",
        "description",
        "partial_on_complete_genome:Boolean",
        "missing_start:Boolean",
        "missing_stop:Boolean",
        "internal_stop:Boolean",
        "partial_in_the_middle_of_a_contig:Boolean",
        "missing_N_terminus:Boolean",
        "missing_C_terminus:Boolean",
        "frameshifted:Boolean",
        "too_short_partial_abutting_assembly_gap:Boolean",
        "incomplete:Boolean",
    ]
    property_specification = {
        "external_id": str,
        "locus_tag": str,
        "start": int,
        "end": int,
        "strand": int,
        "description": str,
        "partial_on_complete_genome": bool,
        "missing_start": bool,
        "missing_stop": bool,
        "internal_stop": bool,
        "partial_in_the_middle_of_a_contig": bool,
        "missing_N_terminus": bool,
        "missing_C_terminus": bool,
        "frameshifted": bool,
        "too_short_partial_abutting_assembly_gap": bool,
        "incomplete": bool,
    }
    start_class = Nodes.NUCLEOTIDE
    end_class = Nodes.PROTEIN


class TAXON_PARENT(Relationship):
    """Represents the relationship between a taxon and its parent taxon"""

    neo4j_label = "TAXON_PARENT"
    description = "Represents the relationship between a taxon and its parent taxon"
    header_filename = "taxid_to_taxid.header"
    target_subdirectory = "taxdump_process"
    target_extension = "taxid_to_taxid"
    header = [":START_ID(taxid)", ":END_ID(taxid)"]
    start_class = Nodes.TAXID
    end_class = Nodes.TAXID


class GO_ANN(Relationship):
    """Represents the relationship between a TIGRFAM HMM and a GO term"""

    neo4j_label = "GO_ANN"
    description = "Represents the relationship between a TIGRFAM HMM and a GO term"
    header_filename = "tigrfam_to_go.header"
    target_subdirectory = "tigrfam_info"
    target_extension = "tigrfam_to_go"
    header = [":START_ID(hmm_source)", ":END_ID(goterm)"]
    start_class = Nodes.HMM_SOURCE
    end_class = Nodes.GOTERM


class PROTEIN_TO_GO(Relationship):
    """Represents the relationship between a protein and a GO term"""

    neo4j_label = "PROTEIN_TO_GO"
    description = "Represents the relationship between a protein and a GO term"
    header_filename = "protein_to_go.header"
    target_subdirectory = "protein_info"
    target_extension = "protein_to_go"
    header = [":START_ID(protein)", ":END_ID(goterm)"]
    start_class = Nodes.PROTEIN
    end_class = Nodes.GOTERM


class GOTERM_RELS(Relationship):
    """Represents the relationship between GO terms"""

    neo4j_label = "GOTERM_RELS"
    description = "Represents the relationship between GO terms"
    multilabel = True
    header_filename = "go_to_go.header"
    target_subdirectory = "goterms"
    target_extension = "goterm_edgelist"
    header = [":START_ID(goterm)", ":END_ID(goterm)", ":TYPE"]
    start_class = Nodes.GOTERM
    end_class = Nodes.GOTERM


class ROLE_ANN(Relationship):
    """Represents the relationship between a TIGRFAM HMM and a role"""

    neo4j_label = "ROLE_ANN"
    description = "Represents the relationship between a TIGRFAM HMM and a role"
    header_filename = "tigrfam_to_role.header"
    target_subdirectory = "tigrfam_info"
    target_extension = "tigrfam_to_role"
    header = [":START_ID(hmm_source)", ":END_ID(tigrfam_role)"]
    start_class = Nodes.HMM_SOURCE
    end_class = Nodes.TIGRFAM_ROLE


class MAINROLE_ANN(Relationship):
    """Represents the relationship between a TIGRFAM role and a main role"""

    neo4j_label = "MAINROLE_ANN"
    description = "Represents the relationship between a TIGRFAM role and a main role"
    header_filename = "tigrfamrole_to_mainrole.header"
    target_subdirectory = "tigrfam_info"
    target_extension = "tigrfamrole_to_mainrole"
    header = [":START_ID(tigrfam_role)", ":END_ID(tigrfam_mainrole)"]
    start_class = Nodes.TIGRFAM_ROLE
    end_class = Nodes.TIGRFAM_MAINROLE


class SUBROLE_ANN(Relationship):
    """Represents the relationship between a TIGRFAM role and a subrole"""

    neo4j_label = "SUBROLE_ANN"
    description = "Represents the relationship between a TIGRFAM role and a subrole"
    header_filename = "tigrfamrole_to_subrole.header"
    target_subdirectory = "tigrfam_info"
    target_extension = "tigrfamrole_to_subrole"
    header = [":START_ID(tigrfam_role)", ":END_ID(tigrfam_subrole)"]
    start_class = Nodes.TIGRFAM_ROLE
    end_class = Nodes.TIGRFAM_SUBROLE


class IS_TAXON(Relationship):
    """Represents the relationship between an assembly and a taxon"""

    neo4j_label = "IS_TAXON"
    description = "Represents the relationship between an assembly and a taxon"
    header_filename = "assembly_to_taxid.header"
    target_subdirectory = "genomic_info"
    target_extension = "assembly_to_taxid"
    header = [":START_ID(assembly)", ":END_ID(taxid)"]
    start_class = Nodes.ASSEMBLY
    end_class = Nodes.TAXID


class BLASTP(Relationship):
    """Represents the relationship between two proteins calculated with DIAMOND BLASTp"""

    neo4j_label = "BLASTP"
    description = "Represents the relationship between two proteins calculated with DIAMOND BLASTp"
    header_filename = "blastp.header"
    target_subdirectory = "diamond_blastp"
    target_extension = "blast6"
    header = [
        ":START_ID(protein)",
        ":END_ID(protein)",
        "pident:Float",
        "length:Long",
        "mismatch:Long",
        "gapopen:Long",
        "qstart:Long",
        "qend:Long",
        "sstart:Long",
        "send:Long",
        "evalue:Float",
        "bitscore:Float",
        "qcovhsp:Float",
    ]
    start_class = Nodes.PROTEIN
    end_class = Nodes.PROTEIN


class MMSEQS2(Relationship):
    """Represents the relationship between two proteins calculated with MMseqs2"""

    neo4j_label = "MMSEQS2"
    description = (
        "Represents the relationship between two proteins calculated with MMseqs2"
    )
    header_filename = "mmseqs2.header"
    target_subdirectory = "mmseqs2_cluster"
    target_extension = "mmseqs2_results_cluster.tsv"
    header = [":START_ID(protein)", ":END_ID(protein)", ":TYPE"]
    start_class = Nodes.PROTEIN
    end_class = Nodes.PROTEIN


class SOURCE_DB(Relationship):
    """Represents the relationship between a HMM and HMM source database(s)"""

    neo4j_label = "SOURCE_DB"
    description = "Represents the relationship between a HMM and HMM source database(s)"
    header_filename = "hmm_source_relationships.header"
    target_subdirectory = "hmm_info"
    target_extension = ".hmminfo"
    header = [
        ":END_ID(hmm_source)",
        ":IGNORE",
        ":IGNORE",
        ":IGNORE",
        ":IGNORE",
        ":IGNORE",
        ":IGNORE",
        ":IGNORE",
        ":IGNORE",
        ":START_ID(hmm)",
        ":IGNORE",
        ":IGNORE",
        ":IGNORE",
        ":IGNORE",
        ":IGNORE",
        ":IGNORE",
        ":IGNORE",
    ]
    start_class = Nodes.HMM
    end_class = Nodes.HMM_SOURCE



import argparse
import csv
import gzip
from pathlib import Path

from socialgene.base.molbio import Domain
from socialgene.base.socialgene import SocialGene
from socialgene.config import env_vars
from socialgene.utils.logging import log

parser = argparse.ArgumentParser(description="Create *.locus_to_protein a genbank file")

parser.add_argument(
    "--input",
    help="One input domtblout file. If a directory, must also use --glob. **Must** be an HMMSEARCH (**not** hmmscan) domtblout file",
    required=True,
    type=str,
    nargs="+",
)
parser.add_argument(
    "--glob",
    help="Glob to input domtblout files. **Must** be an HMMSEARCH (**not** hmmscan) domtblout file ",
    required=False,
    type=str,
)
parser.add_argument(
    "--outpath",
    help="Output filepath",
    required=True,
)
parser.add_argument(
    "--ievaluefilter",
    help="If true, will filter hmmsearch results by env['HMMSEARCH_IEVALUE'] default limit set in common_parameters.env.",
    required=False,
    default=None,
    action=argparse.BooleanOptionalAction,
)


def process_domtblout_file(domtblout_file, ievaluefilter, domain_counter):
    socialgene_object = SocialGene()
    for i in socialgene_object._parse_domtblout(
        input_path=domtblout_file, hmmsearch_or_hmmscan="hmmsearch"
    ):
        if ievaluefilter:
            filter_value = float(env_vars["HMMSEARCH_IEVALUE"])
        else:
            # if not set, use infinity (keep everything)
            filter_value = float("inf")
        if float(i["i_evalue"]) <= filter_value:
            domain_obj = Domain(**i, exponentialized=True)
            _temp = [i["external_id"]]
            _temp.extend(list(domain_obj.tsv_attributes().values()))
            domain_counter += 1
            yield _temp


def main(args=None):
    args = parser.parse_args(args)
    hmm_files = []
    for i in (Path(i) for i in list(args.input)):
        if i.is_dir():
            if args.glob:
                for ii in list(Path(i).glob(args.glob)):
                    hmm_files.append(ii)
        elif i.is_file():
            hmm_files.append(i)
    if not hmm_files:
        if args.glob:
            log.warning(
                f"No domtblout files found in {args.input} with glob {args.glob}"
            )
        else:
            log.warning(f"No domtblout files found in {args.input}")
    else:
        domain_counter = 0
        log.info(f"Socialgene variables: \n{env_vars}")
        log.info(f"sg_process_domtblout will write data to {args.outpath}")
        with gzip.open(args.outpath, "wt", compresslevel=3) as f:
            tsv_writer = csv.writer(f, delimiter="\t")
            for domtblout_file in hmm_files:
                for i in process_domtblout_file(
                    domtblout_file,
                    ievaluefilter=args.ievaluefilter,
                    domain_counter=domain_counter,
                ):
                    tsv_writer.writerow(i)
                    domain_counter += 1
        log.info(f"Wrote {str(domain_counter)} domains to {args.outpath}")


if __name__ == "__main__":
    main()



import subprocess
from pathlib import Path
from shutil import which
from typing import Generator

import socialgene.utils.file_handling as fh
from socialgene.config import env_vars
from socialgene.parsers.hmmmodel import HmmParse
from socialgene.utils.logging import log
from socialgene.utils.run_subprocess import run_subprocess

HMMPRESS_OUTPUT_SUFFIXES = ("h3f", "h3i", "h3m", "h3p")


class HMMER:
    """
    The `HMMER` class provides methods for running HMMER programs, such as `hmmpress` and `hmmscan`, and
    checking for the presence of HMMER installation.
    """

    def __init__(self, hmm_filepath=None):
        self._check_hmmer_exists()
        self.input_path = Path(hmm_filepath)
        self.decompressed_hmm_path = None

    @staticmethod
    def _check_hmmer_exists():
        """
        The function checks if HMMER is installed and provides the location if found, otherwise it
        displays a warning message.
        """
        if which("hmmsearch"):
            log.info(
                f"Will be using HMMER programs located in: {Path(which('hmmsearch')).parents[0]}"
            )
        else:
            log.warning(
                "Couldn't find HMMER, make sure it is installed and on your PATH"
            )

    @property
    def has_cutoffs(self):
        # Check if the first model in the file has GAthering cutoffs
        path_to_test = None
        if Path(self.input_path).exists():
            path_to_test = self.input_path
        elif Path(self.decompressed_hmm_path).exists():
            path_to_test = self.decompressed_hmm_path
        else:
            raise FileNotFoundError(
                f"Couldn't find hmm file: {self.input_path} or {self.decompressed_hmm_path}"
            )
        temp = HmmParse()
        temp_read = temp.read_model_generator(filepath=path_to_test)
        first_model = temp_read.__next__()
        return first_model.has_cutoffs

    def _decompress_hmm_file_if_needed(
        self,
    ):
        # hmmpress requires a decompressed hmm model file
        # This function checks if the decompressed file exists

        is_compressed = fh.is_compressed(self.input_path).name == "gzip"
        if is_compressed and self.input_path.suffix != ".gz":
            raise ValueError(
                f"Compressed hmm file must have a '.gz' extension: {self.input_path}"
            )
        if is_compressed and self.input_path.with_suffix("").exists():
            log.debug(
                f"Decompressed hmm file already exists: {self.input_path.with_suffix('')}"
            )
            return
        if is_compressed:
            self.decompressed_hmm_path = fh.gunzip(self.input_path)
            log.debug(
                f"Decompressed hmm file: {self.input_path} -> {self.decompressed_hmm_path}"
            )
        else:
            self.decompressed_hmm_path = self.input_path

    def _append_hmmpress_suffixes(self) -> Generator[str, None, None]:
        """Appends each of the hmmpress file extensions to the given filepath
        Returns:
            Generator: new paths of expected hmmpress output files
        """
        return (f"{self.decompressed_hmm_path}.{i}" for i in HMMPRESS_OUTPUT_SUFFIXES)

    def _check_hmmpress_files_exist(self) -> bool:
        """Checks if all the expected files are present after running hmmpress
        Returns:
            bool: True if all expected files are present, False otherwise
        """
        if all((Path(i).exists() for i in self._append_hmmpress_suffixes())):
            return True
        else:
            log.debug("Not all expected hmmpress files found")
            log.debug(
                f"Missing: {(i for i in self._append_hmmpress_suffixes() if not i.exists())}"
            )
            return False

    def hmmpress(self, force=False):
        """
        The function `hmmpress` checks if the HMM file has the correct extension, determines the file
        path, and skips the hmmpress step if the necessary files already exist.

        Args:
          hmm_path: The path to the HMM file that needs to be processed.
          force: The `force` parameter is a boolean flag that determines whether to force the execution
        of the `hmmpress` function or not. If `force` is set to `True`, the function will be executed
        regardless of whether the necessary files already exist. Defaults to False

        Returns:
          nothing (None).
        """
        self._decompress_hmm_file_if_needed()
        # `self._decompress_hmm_file_if_needed()` populates `self.decompressed_hmm_path`
        if not force and self._check_hmmpress_files_exist():
            log.info(
                "hmmpress outputs found and hmmpress(force=False), skipping hmmpress"
            )
            return
        if force:
            command_list = ["hmmpress", "-f", str(self.decompressed_hmm_path)]
        else:
            command_list = ["hmmpress", str(self.decompressed_hmm_path)]
        command_list = [str(i) for i in command_list]
        run_subprocess(
            command_list=command_list,
        )
        if not self._check_hmmpress_files_exist():
            raise FileNotFoundError("Didn't find expected files after running hmmpress")

    def hmmscan(
        self,
        domtblout_path: str,
        fasta_path: str,
        input: str = None,
        f1: float = env_vars["HMMSEARCH_F1"],
        f2: float = env_vars["HMMSEARCH_F2"],
        f3: float = env_vars["HMMSEARCH_F3"],
        dom_e=env_vars["HMMSEARCH_DOME"],
        inc_e=env_vars["HMMSEARCH_INCE"],
        incdom_e=env_vars["HMMSEARCH_INCDOME"],
        e=env_vars["HMMSEARCH_E"],
        z=env_vars["HMMSEARCH_Z"],
        seed: int = env_vars["HMMSEARCH_SEED"],
        cpus: int = 1,
        overwrite=False,
    ) -> None:
        """
        The `hmmscan` function runs HMMER's hmmscan tool to search for protein domains in a given FASTA
        file using a specified HMM profile, and saves the results in a specified output file.

        Args:
          fasta_path (str): The path to the FASTA file that contains the sequences to be searched
        against the HMM profiles.
          input (str): use to pass fasta as stdin
          hmmpath (str): The `hmmpath` parameter is the path to the HMM file that will be used
        for the hmmscan.
          domtblout_path (str): The `domtblout_path` parameter is the path where the output file from
        HMMER's hmmscan will be saved. This file will contain the domain hits found in the input
        sequences.
          f1 (float): The parameter `f1` is a float value representing the gathering threshold for the
        first domain architecture. It is used in the HMMER's hmmscan command to set the threshold for
        the first domain architecture. The default value is obtained from the `env_vars` dictionary with
        the key "H
          f2 (float): The parameter `f2` in the `hmmscan` method is a float that represents the
        gathering threshold for the second domain architecture. It is used as an optional argument in
        the `hmmscan` command.
          f3 (float): The parameter `f3` is a float value that represents the threshold for the third
        stage of the HMMER's hmmscan algorithm. It is used to control the trade-off between sensitivity
        and speed during the search. A higher value of `f3` will increase the speed of the search
          dom_e: The `dom_e` parameter is the E-value threshold for reporting domain hits. It is used to
        control the significance threshold for reporting individual domains in the output.
          inc_e: The `inc_e` parameter is the inclusion threshold for the E-value. It is used to
        determine if a hit is significant enough to be included in the output. Hits with E-values below
        the inclusion threshold will be reported.
          incdom_e: The `incdom_e` parameter is the inclusion threshold for domain E-values in HMMER's
        hmmscan. It is used to control the sensitivity of the search by setting a threshold for the
        significance of domain hits. Any domain hit with an E-value below the `incdom_e` threshold
          e: The parameter `e` in the `hmmscan` method is the E-value threshold for reporting domains.
        It is a floating-point value that determines the significance threshold for the reported
        matches. A lower E-value indicates a more significant match.
          z: The parameter `z` in the `hmmscan` method is used to set the reporting threshold for the
        gathering score. The gathering score is a measure of the significance of a match between a
        sequence and a profile hidden Markov model (HMM). The `z` parameter specifies the number of
          seed (int): The `seed` parameter in the `hmmscan` method is an optional integer parameter that
        specifies the random number seed for the HMMER's hmmscan algorithm.
          cpus (int): The `cpus` parameter specifies the number of CPUs (or cores) to be used for
        running the hmmscan process. Defaults to 1
          overwrite: The `overwrite` parameter is a boolean flag that determines whether to overwrite an
        existing file at the `domtblout_path` location. If `overwrite` is set to `False` and a file
        already exists at `domtblout_path`, a `FileExistsError` will be raised. Defaults to False
        """
        self.hmmpress()

        if self.has_cutoffs:
            # base on the first model in the file
            use_ga_cutoffs = True
        else:
            use_ga_cutoffs = False

        if input:
            # set hmmscan arg to read from stdin
            fasta_path = "-"

        domtblout_path = Path(domtblout_path)
        if domtblout_path.exists() and not overwrite:
            raise FileExistsError(f"File already exists at {domtblout_path}")
        c1 = [
            "hmmscan",
            "--seed",
            int(seed),
            "--cpu",
            int(cpus),
            "-Z",
            z,
        ]

        c3 = [
            "--F1",
            f1,
            "--F2",
            f2,
            "--F3",
            f3,
            "--domtblout",
            domtblout_path,
            self.decompressed_hmm_path,
            fasta_path,
        ]
        if use_ga_cutoffs:
            c2 = ["--cut_ga"]
        else:
            c2 = [
                "-E",
                e,
                "--incE",
                inc_e,
                "--incdomE",
                incdom_e,
                "--domE",
                dom_e,
            ]
        command_list = c1 + c2 + c3
        command_list = [str(i) for i in command_list]
        run_subprocess(
            command_list=command_list,
            input=input,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            shell=False,
            capture_output=False,
        )



import argparse

from socialgene.neo4j.schema.socialgene_modules import SocialgeneModules
from socialgene.utils.logging import log

parser = argparse.ArgumentParser(
    description="Export header files for Neo4j admin import"
)

parser.add_argument(
    "--outdir",
    help="Output directory filepath",
    required=True,
)
parser.add_argument(
    "--sg_modules",
    help="Modules for data import; current choices: ['base', 'blastp', 'hmms', 'mmseqs2', 'ncbi_taxonomy', 'paired_omics']",
    required=True,
    nargs="+",
)
parser.add_argument(
    "--include_sequences",
    help="Should sequences be included in the database?",
    required=False,
    default=False,
    action=argparse.BooleanOptionalAction,
)


def main():
    args = parser.parse_args()
    log.info(args.sg_modules)
    sg_mod = SocialgeneModules(include_sequences=args.include_sequences)
    sg_mod.add_modules(args.sg_modules)
    sg_mod.write_neo4j_headers(outdir=args.outdir)


if __name__ == "__main__":
    main()



# import sqlite3
# from pathlib import Path

# import pandas as pd
# from rich.progress import Progress

# # this assumes that the protein IDs are all unique
# # because this is only used for refseq/genbank data that should be true


# def create_db(tsv_list, outpath, inmem=True):
#     sqliteout = Path(outpath)
#     # files = list(Path("/home/chase/Downloads/a").glob("*.tsv"))
#     if inmem:
#         # build database in-memory then write out to disk at the end
#         conn = sqlite3.connect(":memory:")
#     else:
#         # build database on disk
#         conn = sqlite3.connect(sqliteout)
#     c = conn.cursor()
#     c.execute("""DROP TABLE IF EXISTS protmap""")
#     conn.execute(
#         """
#         CREATE TABLE protmap (
#             hash CHAR(32) NOT NULL,
#             accession VARCHAR(32) NOT NULL,
#             UNIQUE(accession)
#         );
#         """
#     )
#     c.execute(
#         """
#     CREATE INDEX accession_index
#     ON protmap (accession);
#     """
#     )
#     c.execute("""PRAGMA synchronous = OFF""")
#     c.execute("""PRAGMA journal_mode = OFF""")
#     c.execute("""PRAGMA TEMP_STORE = MEMORY""")
#     c.execute("""PRAGMA cache_size = 1000000""")
#     c.execute("""PRAGMA PAGE_SIZE = 1000000""")
#     c.execute("""PRAGMA ignore_check_constraints = true""")
#     with Progress(transient=True) as progress:
#         task = progress.add_task("Progress...", total=len(tsv_list))
#         for i in tsv_list:
#             progress.update(task, advance=1)
#             pd.read_csv(
#                 i, sep="\t", header=None, names=["sha512t24u", "accession"]
#             ).to_sql("protmap", con=conn, if_exists="append", index=False)
#     if inmem:
#         # save in-memory sqlite database to disk
#         db_disk = sqlite3.connect(sqliteout)
#         conn.backup(db_disk)



import os
import tempfile
from pathlib import Path

import pandas as pd

from socialgene.base.molbio import Protein
from socialgene.compare_proteins.base import BlastTab, BlastTab_COLUMNS
from socialgene.utils.logging import log
from socialgene.utils.run_subprocess import run_subprocess


class DiamondBlastp(BlastTab):
    def __init__(self):
        super().__init__()
        self.temp_db_path = None
        self.df = None
        self.name = "Diamond BLASTp"

    def make_db(
        self,
        fasta_string: str,
        db_path=None,
        threads=1,
        **kwargs,
    ):
        db_path = Path(db_path).with_suffix(".dmnd")
        # run diamond makedb on fasta_file allow passing kwargs to diamond
        command_list = [
            "diamond",
            "makedb",
            "--db",
            str(db_path),
            "--threads",
            str(threads),
            "--verbose",
        ]
        mes = run_subprocess(
            command_list=command_list,
            input=fasta_string,
            text=True,
            capture_output=True,
            status=False,
            **kwargs,
        )
        log.debug(mes)
        if not Path(db_path).exists():
            raise FileNotFoundError(
                f"No database from DiamondBlastp make_db found at {db_path}"
            )

    @staticmethod
    def run(fasta_path, db_path=None, threads=1, argstring="", **kwargs):
        # run diamond blastp on query_fasta_string against db_path
        if not Path(db_path).exists():
            raise FileNotFoundError(
                f"No database from Diamond makedb found at {db_path}"
            )
        with tempfile.TemporaryDirectory() as tmpdirname:
            outpath = os.path.join(tmpdirname, "result.m8")
            command_list = [
                "diamond",
                "blastp",
                "--query",
                str(fasta_path),
                "--db",
                str(db_path),
                "--out",
                str(outpath),
                "--threads",
                str(threads),
                "--verbose",
                "--outfmt",
                "6",
                "qseqid",
                "sseqid",
                "pident",
                "length",
                "mismatch",
                "gapopen",
                "qstart",
                "qend",
                "sstart",
                "send",
                "evalue",
                "bitscore",
                "qlen",
                "slen",
            ]
            if argstring:
                argstring = [str(i) for i in argstring.split()]
                command_list.extend(argstring)
            mes = run_subprocess(
                command_list=command_list,
                check=False,
                shell=False,
                capture_output=True,
                status=False,
            )
            log.debug(mes)
            if not os.path.exists(outpath):
                log.warning("No output from Diamond search")
            else:
                # sorted for reproducibility
                return (
                    pd.read_csv(
                        outpath,
                        sep="\t",
                        header=None,
                        names=BlastTab_COLUMNS,
                        dtype=BlastTab_COLUMNS,
                    )
                    .sort_values(["query", "score"], ascending=False)
                    .reset_index(drop=True)
                )

    def compare_proteins(
        self,
        queries,
        targets,
        threads=1,
        argstring="--fast --no-self-hits --max-hsps 1",
    ):
        # loop through protein list and write to temporary fasta file
        if isinstance(queries, Protein):
            queries = [queries]
        if isinstance(targets, Protein):
            targets = [targets]
        if not all([isinstance(i, Protein) for i in queries]):
            raise TypeError("queries must be a list of Protein objects")
        if not all([isinstance(i, Protein) for i in targets]):
            raise TypeError("targets must be a list of Protein objects")
        with tempfile.TemporaryDirectory() as tmpdirname:
            query_fasta_path = Path(tmpdirname, "queries.faa")
            target_dmnd_path = Path(tmpdirname, "targets.dmnd")
            with open(query_fasta_path, "w") as h:
                for protein in queries:
                    h.write(protein.fasta_string_defline_uid)
            self.make_db(
                fasta_string="".join([i.fasta_string_defline_uid for i in targets]),
                db_path=target_dmnd_path,
            )
            return self.run(
                fasta_path=query_fasta_path,
                db_path=target_dmnd_path,
                threads=threads,
                argstring=argstring,
            )



from collections import OrderedDict

from socialgene.utils.logging import log


class LocusAssemblyMetadata:
    # fmt: off
    __slots__= ["altitude","bio_material","bioproject","biosample","cell_line","cell_type","chromosome","clone","clone_lib","collected_by","collection_date","country","cultivar","culture_collection","db_xref","dev_stage","ecotype","environmental_sample","focus","germline","haplogroup","haplotype","host","identified_by","isolate","isolation_source","lab_host","lat_lon","macronuclear","map","mating_type","metagenome_source","mol_type","note","organelle","organism","pcr_primers","plasmid","pop_variant","proviral","rearranged","segment","serotype","serovar","sex","specimen_voucher","strain","sub_clone","submitter_seqid","sub_species","sub_strain","tissue_lib","tissue_type","transgenic","type_material","variety"]  # noqa: E231,E225
    # fmt: on

    def __init__(self, **kwargs) -> None:
        self.update(kwargs)

    def all_attributes(self):
        return OrderedDict({i: self.__getitem__(i) for i in self.__slots__})

    @property
    def values(self):
        return self.all_attributes().values()

    def __getitem__(self, item):
        try:
            return self.__getattribute__(item)
        except Exception as e:
            log.debug(e)
            return None

    # add a function that takes a dict as input and sets any attributes in the dict keys
    def update(self, d: dict):
        for k, v in d.items():
            temp_v = v
            if isinstance(v, list):
                # if info is a list, collapse into a single string
                if len(v) > 1:
                    temp_v = ";".join(v)
                else:
                    temp_v = v[0]
            if not isinstance(temp_v, str):
                temp_v = str(temp_v)
            temp_v.replace("\t", "")
            temp_v.replace("\n", "")
            temp_v.replace("\r", "")
            temp_v.strip()
            if k in self.__slots__:
                setattr(self, k, temp_v)



"""http://classyfire.wishartlab.com"""

import re
from io import BytesIO
from zipfile import ZipFile

import requests

from socialgene.addons.chebi.nr import ChebiNode
from socialgene.addons.classyfire.nr import (
    ClassyFireIsA,
    ClassyFireNode,
    ClassyFireSynonym,
)
from socialgene.utils.logging import log

OBO_URL = (
    "http://classyfire.wishartlab.com/system/downloads/1_0/chemont/ChemOnt_2_1.obo.zip"
)


class ClassyFireEntry:
    __slots__ = ["uid", "name", "definition", "chebi", "is_a"]

    def __init__(self) -> None:
        self.uid = None
        self.name = None
        self.definition = None
        self.chebi = set()
        self.is_a = None

    def assign(self, line):
        if line.startswith("id:"):
            self._add_id(line)
        elif line.startswith("name:"):
            self._add_name(line)
        elif line.startswith("def:"):
            self._add_definition(line)
        elif line.startswith("synonym:"):
            self._add_chebilist(line)
        elif line.startswith("is_a:"):
            self._add_is_a(line)

    def _add_id(self, x):
        self.uid = x.removeprefix("id: CHEMONTID:").strip()
        self.uid = int(self.uid)

    def _add_name(self, x):
        self.name = x.removeprefix("name:").strip()

    def _add_definition(self, x):
        self.definition = x.split('"')[1].strip()

    def _add_chebilist(self, x):
        matched = re.search("CHEBI:[0-9]+", x)
        if matched:
            self.chebi.add(matched.group().removeprefix("CHEBI:").strip())

    def _add_is_a(self, x):
        matched = re.search("CHEMONTID:[0-9]+", x)
        if matched:
            self.is_a = matched.group().removeprefix("CHEMONTID:").strip()
            self.is_a = int(self.is_a)


class ClassyFire:
    def __init__(self) -> None:
        self.entries = self.download()

    @staticmethod
    def download(url=OBO_URL):
        response = requests.get(url, stream=True)
        obs = []
        log.info(f"Downloading and parsing data from:\n\t{OBO_URL}")
        with ZipFile(BytesIO(response.content)).open("ChemOnt_2_1.obo") as h:
            term_open = False
            for line in h:
                line = line.decode("utf-8")
                if line.startswith("[Term]"):
                    term_open = True
                    obj = ClassyFireEntry()
                if not term_open:
                    continue
                obj.assign(line)
                if line == "\n":
                    term_open = False
                    obs.append(obj)
        return obs

    def add_classyfire_nodes(self, create=False):
        b = [
            ClassyFireNode(
                properties={"uid": i.uid, "name": i.name, "definition": i.definition}
            )
            for i in self.entries
        ]
        ClassyFireNode.add_multiple_to_neo4j(b, create=create)

    def add_classyfire_isa_relationships(self, create=False):
        b = [
            ClassyFireIsA(
                start=ClassyFireNode(properties={"uid": i.uid}),
                end=ClassyFireNode(properties={"uid": i.is_a}),
                create=create,
            )
            for i in self.entries
            if i.is_a
        ]
        ClassyFireIsA.add_multiple_to_neo4j(b, create=create)

    def add_classyfire_synonym_relationships(self, create=False):
        zz = []
        chebi_ids = set()
        for i in self.entries:
            for ii in i.chebi:
                chebi_ids.add(ii)
                zz.append(
                    ClassyFireSynonym(
                        start=ClassyFireNode(properties={"uid": i.uid}),
                        end=ChebiNode(properties={"uid": int(ii)}),
                        create=create,
                    )
                )
        ChebiNode.add_multiple_to_neo4j(
            [ChebiNode(properties={"uid": int(i)}) for i in chebi_ids], create=create
        )
        ClassyFireSynonym.add_multiple_to_neo4j(zz, create=create)



from textdistance import jaccard, levenshtein

from socialgene.base.molbio import Protein

# Note: there is technically no query/target as the comparison methods are order-independent
# this is just to keep things consistent between tools
# _mod_score_tupler = namedtuple(
#     "protein_comparison_modscore",
#     (
#         "query",
#         "target",
#         "query_n_domains",
#         "target_n_domains",
#         "levenshtein",
#         "jaccard",
#         "mod_score",
#     ),
# )


class _mod_score_tupler:
    __slots__ = (
        "query",
        "target",
        "query_n_domains",
        "target_n_domains",
        "levenshtein",
        "jaccard",
        "mod_score",
    )

    def __init__(
        self,
        query: str,
        target: str,
        query_n_domains: int,
        target_n_domains: int,
        levenshtein: int,
        jaccard: int,
        mod_score: int,
    ):
        for i in self.__slots__:
            setattr(self, i, locals()[i])
        if not isinstance(query, Protein):
            raise TypeError(f"query type: {type(query)}")
        if not isinstance(target, Protein):
            raise TypeError(f"target type: {type(target)}")

    def __dict__(self):
        return {
            "query": self.query.uid,
            "target": self.target.uid,
            "query_n_domains": self.query_n_domains,
            "target_n_domains": self.target_n_domains,
            "levenshtein": self.levenshtein,
            "jaccard": self.jaccard,
            "mod_score": self.mod_score,
        }

    def to_dict(self):
        return self.__dict__()

    def __repr__(self):
        return f"_mod_score_tupler(query=Protein(uid='{self.query.uid}'), target=Protein(uid='{self.target.uid}'), query_n_domains={self.query_n_domains}, target_n_domains={self.target_n_domains}, levenshtein={self.levenshtein}, jaccard={self.jaccard}, mod_score={self.mod_score})"


def mod_score(p1, p2):
    """Combine Edit and Composition Distances (Levenshtein and Jaccard)
    Important: If protein sequence hashes are identical then a perfect score is returned without comparing domains.
    Args:
        p1 (Protein): SocialGene Protein Class
        p1 (Protein): SocialGene Protein Class
    Returns:
        dict: {ProteinClass, ProteinClass, levenshtein, jaccard, mod_score}; mod_score -> 2 = Perfectly similar; otherwise (1/Levenshtein + Jaccard)
    """
    # If either protein contains no HMM annotations,
    # return a mod score with the worst scores possible
    if not isinstance(p1, Protein) and not isinstance(p2, Protein):
        raise TypeError(f"p1 type: {type(p1)}; p2 type: {type(p2)}")
    length_input_list_1 = len(p1.domains)
    length_input_list_2 = len(p2.domains)
    if p1 == p2:
        return _mod_score_tupler(
            p1,
            p2,
            length_input_list_1,
            length_input_list_2,
            round(1, 2),
            round(1, 2),
            round(1.5, 2),
        )
    if not p1.domains or not p2.domains:
        # If either protein contains no HMM annotations,
        # return a mod score with the worst scores possible
        return _mod_score_tupler(
            p1,
            p2,
            length_input_list_1,
            length_input_list_2,
            round(0, 2),
            round(0, 2),
            round(0, 2),
        )
    # levenshtein_score: 0=most similar; increases based on number of edits
    levenshtein_score = levenshtein(p1.domain_vector, p2.domain_vector)
    # convert int64 to int
    levenshtein_score = int(levenshtein_score)
    # jaccard_score: 1= most similar; 0=most dissimilar
    jaccard_score = jaccard(set(p1.domain_vector), set(p2.domain_vector))
    # change levenshtein so 0=worst, 1 = best
    max_edit = max(length_input_list_1, length_input_list_2)
    if max_edit == 0:
        # Neither protein had HMMs (case is actually handled above (if not p1.domains or not p2.domains:))
        mod_score_value = 0
        mod_levenshtein_score = 0
    elif length_input_list_1 == 0 or length_input_list_2 == 0:
        # One protein didn't have HMMs
        mod_score_value = 0
        mod_levenshtein_score = 0
    else:
        mod_levenshtein_score = 1 - (levenshtein_score / max_edit)
        if jaccard_score == 0:
            mod_score_value = 0
        else:
            mod_score_value = (jaccard_score * 0.5) + mod_levenshtein_score

    return _mod_score_tupler(
        p1,
        p2,
        length_input_list_1,
        length_input_list_2,
        round(mod_levenshtein_score, 2),
        round(jaccard_score, 2),
        round(mod_score_value, 2),
    )



# parse gff file into socialgene object
from pathlib import Path

from Bio import SeqIO

import socialgene.utils.file_handling as fh
from socialgene.utils.logging import log


class GFFParserMixin:
    def __init__(self):
        pass

    def _check_is_gff(self, input_path: str):
        if fh.guess_filetype(input_path) != "gff":
            raise ValueError("Input file is not a GFF file")

    def _has_fasta(self, input_path: str):
        with fh.open_read(input_path) as f:
            for line in f:
                if line.startswith("##FASTA"):
                    return True
        return False

    def _fasta_biopython(self, input_path: str):
        with fh.open_read(input_path) as file:
            for line in file:
                if line.startswith("##FASTA"):
                    break
            # read the remaining lines into fasta dictionary using biopython
            fasta_dict = SeqIO.to_dict(SeqIO.parse(file, "fasta"))
            if not fasta_dict:
                log.warning("No sequences found in FASTA section")
        return fasta_dict

    def parse_gff_file(self, input_path: str, keep_sequence: bool = True, **kwargs):
        # name of file without extension
        assembly_uid = Path(input_path).name
        assembly_uid = assembly_uid.split(".gff")[0]
        self._check_is_gff(input_path)
        self._has_fasta(input_path)
        nucleotide_sequence_dict = self._fasta_biopython(input_path)
        self.add_assembly(assembly_uid)
        with fh.open_read(input_path) as file:
            for line in file:
                if line.startswith("#") or line.startswith(">"):
                    continue
                parts = line.strip().split("\t")
                if len(parts) != 9:
                    continue
                (
                    seq_id,
                    source,
                    feature_type,
                    start,
                    end,
                    score,
                    strand,
                    phase,
                    attributes,
                ) = parts
                if feature_type not in ["protein", "CDS", "pseudogene"]:
                    continue
                # get the translated sequence using biopython's translate after extracting the sequence
                # from the fasta file
                sequence = nucleotide_sequence_dict[seq_id][int(start) - 1 : int(end)]
                if strand == "-":
                    sequence = sequence.reverse_complement()
                sequence = sequence.translate()
                uid = self.add_protein(
                    description=None,
                    external_id=None,
                    sequence=str(sequence.seq),
                    return_uid=True,
                )
                if not keep_sequence:
                    self.proteins[uid].sequence = None
                if strand == "-":
                    strand = -1
                elif strand == "+":
                    strand = 1
                else:
                    strand = 0
                self.assemblies[assembly_uid].add_locus(external_id=seq_id)
                self.assemblies[assembly_uid].loci[seq_id].add_feature(
                    type=feature_type,
                    description=None,
                    external_id=None,
                    uid=uid,
                    start=int(start),
                    end=int(end),
                    strand=strand,
                )



import argparse
import base64
import hashlib

from Bio.SeqUtils import CheckSum

from socialgene.config import env_vars

parser = argparse.ArgumentParser(description="Hash a string/amino acids")
parser.add_argument(
    "--input",
    help="single string to hash",
    required=True,
)


def sha512t24u(input):
    # https://www.ncbi.nlm.nih.gov/pmc/articles/PMC7714221/
    # To standardize to caps-only input, use hash_aminos()
    sha512_digest = hashlib.sha512(bytes(input, "utf8")).digest()[:24]
    sha512t24u = base64.urlsafe_b64encode(sha512_digest).decode("ascii")
    return sha512t24u


def hash_aminos(input, **kwargs):
    # make sure everything is uppercase before hashing
    cleaned = input.upper()
    # remove all whitespace
    cleaned = cleaned.replace(" ", "")
    # remove "*" if it's at the beginning or end of the string but not in the middle
    cleaned = cleaned.strip("*")
    return hasher(input=cleaned, **kwargs)


def use_hashlib(input, algo):
    allow_algos = ("sha512", "sha256", "sha384", "md5", "sha224")
    if algo not in allow_algos:
        # TODO: this should also output the sha512t24u ans crc algos
        raise ValueError(f"algo must be one of: {allow_algos}")
    hasher = getattr(hashlib, algo)
    return hasher(bytes(input, "utf8")).hexdigest()


def hasher(input, algo=None):
    if not algo:
        algo = env_vars["HASHING_ALGORITHM"]
    match algo:
        case "sha512t24u":
            return sha512t24u(input)
        case "crc64":
            return CheckSum.crc64(input).removeprefix("CRC-")
        case "crc32":
            return CheckSum.crc32(input)
        case "seguid":
            return CheckSum.seguid(input)
        case _:
            return use_hashlib(input=input, algo=algo)



from rdkit import Chem, DataStructs, RDLogger
from rdkit.Chem import AllChem, Descriptors, rdMolHash

from socialgene.addons.chemistry.nr import ChemicalCompoundNode
from socialgene.utils.logging import log

# Turn off C++ warnings when in info mode
if log.level != 20:
    RDLogger.DisableLog("rdApp.*")


class ChemicalFragments:
    def __init__(self, **kwargs):
        for i, j in Descriptors._descList:
            if i.startswith("fr_"):
                setattr(self, i, None)

    def add_mol(self, mol):
        for i, j in Descriptors._descList:
            if i.startswith("fr_"):
                setattr(self, i, j(mol))

    def to_dict(self):
        temp = {k: v for k, v in self.__dict__.items() if k.startswith("fr_") and v > 0}
        temp = {i.removeprefix("fr_"): j for i, j in temp.items()}
        return temp


def morgan_fingerprint(rdkitmol, radius=2, nBits=2048):
    return AllChem.GetMorganFingerprintAsBitVect(
        rdkitmol, useChirality=True, radius=radius, nBits=nBits, bitInfo={}
    )


class ChemicalCompound:
    def __init__(self, compound, sanitize=True, **kwargs):
        self.mol = None
        self.parse_compound(compound, sanitize=sanitize)
        self.morgan = morgan_fingerprint(self.mol, **kwargs)
        self.inchi = Chem.MolToInchi(self.mol)

    def parse_compound(self, input, sanitize=True):
        if isinstance(input, Chem.rdchem.Mol):
            self.mol = input
        elif isinstance(input, str):
            method_list = [Chem.MolFromInchi, Chem.MolFromSmiles, Chem.MolFromMolFile]
            for method in method_list:
                log.debug(f"Trying to parse compound with {method.__name__}")
                try:
                    temp = method(input, sanitize=sanitize)
                    if isinstance(temp, Chem.rdchem.Mol):
                        self.mol = temp
                        log.debug(
                            f"Successfully parsed compound with {method.__name__}"
                        )
                        break
                except Exception:
                    continue
        if not isinstance(self.mol, Chem.rdchem.Mol):
            raise ValueError("Wasn't able to parse the compound")

    @property
    def hash_dict(self):
        temp = {
            k: rdMolHash.MolHash(self.mol, v)
            for k, v in rdMolHash.HashFunction.names.items()
        }
        temp["inchi"] = self.inchi
        return temp

    @property
    def base_properties(self):
        out_dict = {}
        for i in (
            "MolWt",
            "HeavyAtomMolWt",
            "ExactMolWt",
            "NumValenceElectrons",
            "NumRadicalElectrons",
            "HeavyAtomCount",
            "NumAliphaticCarbocycles",
            "NumAliphaticHeterocycles",
            "NumAliphaticRings",
            "NumAromaticCarbocycles",
            "NumAromaticHeterocycles",
            "NumAromaticRings",
            "NumHAcceptors",
            "NumHDonors",
            "NumHeteroatoms",
            "NumRotatableBonds",
            "NumSaturatedCarbocycles",
            "NumSaturatedHeterocycles",
            "NumSaturatedRings",
            "RingCount",
            "MolLogP",
            "MolMR",
        ):
            try:
                out_dict[i] = getattr(Descriptors, i)(self.mol)
            except Exception:
                out_dict[i] = None
        return out_dict

    @property
    def fragments(self):
        fr = ChemicalFragments()
        fr.add_mol(self.mol)
        return fr.to_dict()

    @property
    def node(self):
        temp = ChemicalCompoundNode()
        temp.fill_from_dict(self.base_properties | self.hash_dict)
        return temp


class ChemicalCollection:
    def __init__(self) -> None:
        pass

    def compare_morgan(self, a, b):
        return DataStructs.TanimotoSimilarity(self.morgan, self.morgan)



import argparse
import glob
from pathlib import Path
from typing import List

from socialgene.base.socialgene import SocialGene
from socialgene.config import env_vars
from socialgene.utils.logging import log

parser = argparse.ArgumentParser(
    description="Export Socialagene object into TSV files for Nextflow/Neo4j"
)
parser.add_argument(
    "--sequence_files_glob",
    help='A "glob" style input path. See https://docs.python.org/3/library/glob.html for info about globs',
    required=False,
)
parser.add_argument(
    "-f",
    "--sequence_file",
    help="Filepath, may be repeated to add multiple paths",
    nargs="+",
    action="append",
    required=False,
)
parser.add_argument(
    "--outdir",
    help="Output directory filepath",
    required=True,
)
parser.add_argument(
    "--n_fasta_splits",
    help="Number of fasta files to create",
    required=True,
    default=1,
)
parser.add_argument(
    "--include_sequences",
    help="Should sequences be included in the database?",
    required=False,
    default=False,
    action=argparse.BooleanOptionalAction,
)
parser.add_argument(
    "--collect_tables_in_memory",
    help="Should all tables be collected in RAM before writing to disk?",
    required=False,
    default=False,
    action=argparse.BooleanOptionalAction,
)
parser.add_argument(
    "--compression",
    help="Gzip compress output?",
    required=False,
    default=False,
    action=argparse.BooleanOptionalAction,
)
parser.add_argument(
    "--defline_magic",
    help="Parse out fasta deflines with defined structures, currently only works for uniprot (deflines that begin with: sp| or tr|)",
    type=bool,
    required=False,
    default=False,
)


def _writer(socialgene_object, outdir, n_fasta_splits, compression, include_sequences):
    socialgene_object.write_n_fasta(
        outdir=outdir, n_splits=n_fasta_splits, mode="a", compression=compression
    )
    for i in SocialGene._export_table_names:
        socialgene_object.write_table(
            outdir=outdir,
            tablename=i,
            filename=i.removeprefix("table_"),
            mode="a",
            compression=compression,
            include_sequences=include_sequences,
        )


def export_tables(
    outdir: str,
    n_fasta_splits: int,
    collect_tables_in_memory: bool,
    compression: str,
    file_list: List = None,
    sequence_files_glob: str = "",
    include_sequences: bool = False,
    defline_magic: bool = False,
):
    """
    The function `export_tables` takes in various parameters, including an output directory, the number
    of fasta splits, a flag to determine if tables should be collected in memory, a compression method,
    and a list of files or a glob pattern to find files. It then imports the files into a SocialGene
    object and writes the tables to the specified output directory.

    Args:
      outdir (str): The `outdir` parameter is a string that specifies the directory where the exported
    tables will be saved.
      n_fasta_splits (int): The `n_fasta_splits` parameter specifies the number of splits to be made
    when exporting the tables. It is used by the `_writer` function to determine how many splits to
    create when writing the tables to files.
      collect_tables_in_memory (bool): A boolean flag indicating whether the tables should be collected
    in memory before writing them to files. If set to True, the tables will be stored in memory and then
    written to files. If set to False, the tables will be written to files immediately after parsing
    each input file.
      compression (str): The "compression" parameter is a string that specifies the compression format
    to use when exporting the tables. It can be one of the following options: "gzip", "bz2", "lzma", or
    "none".
      file_list (List): A list of file paths to input sequence files.
      sequence_files_glob (str): A string representing a glob pattern used to find sequence files. This
    pattern can include wildcards (*) to match multiple files.
      include_sequences (bool): A boolean flag indicating whether the sequences should be included in the database
      defleine_magic (bool): Parse out fasta deflines with defined structures, currently only works for uniprot (deflines that begin with: sp| or tr|)

    """
    outdir = Path(outdir)
    sequence_files = list()
    # find files using the input glob
    if sequence_files_glob:
        sequence_files.extend(glob.glob(sequence_files_glob))
    # and/or use input file paths
    if file_list:
        sequence_files.extend(file_list)
    sequence_files = (Path(i).resolve() for i in sequence_files)
    if not sequence_files:
        raise IOError("No matching file(s)")
    if collect_tables_in_memory:
        # Import all files into the same SG object
        socialgene_object = SocialGene()
        for i in sequence_files:
            socialgene_object.parse(Path(i), defline_magic=defline_magic)
        _writer(
            socialgene_object=socialgene_object,
            outdir=outdir,
            n_fasta_splits=n_fasta_splits,
            compression=compression,
            include_sequences=include_sequences,
        )
    else:
        for i in sequence_files:
            socialgene_object = SocialGene()
            socialgene_object.parse(Path(i), defline_magic=defline_magic)
            _writer(
                socialgene_object=socialgene_object,
                outdir=outdir,
                n_fasta_splits=n_fasta_splits,
                compression=compression,
                include_sequences=include_sequences,
            )


def main():
    args = parser.parse_args()
    log.info(f"Socialgene variables: \n{env_vars}")
    if args.compression:
        compression = "gzip"
    else:
        compression = None
    export_tables(
        sequence_files_glob=args.sequence_files_glob,
        outdir=args.outdir,
        n_fasta_splits=int(args.n_fasta_splits),
        collect_tables_in_memory=args.collect_tables_in_memory,
        compression=compression,
        include_sequences=args.include_sequences,
        defline_magic=args.defline_magic,
    )


if __name__ == "__main__":
    main()



from itertools import combinations_with_replacement
from multiprocessing import Pool, cpu_count

import pandas as pd

from socialgene.compare_proteins.hmm_scoring import mod_score
from socialgene.neo4j.neo4j import Neo4jQuery
from socialgene.utils.logging import log


def append_or_not(result_list, result, append=False):
    if append:
        result_list.extend(result)
    else:
        result_list = result


class CompareProtein(Neo4jQuery):
    def __init__(
        self,
    ):
        super().__init__()
        self.protein_comparison = []

    def _calculate_mod_score_from_protein_class(
        self,
        protein_1=None,
        protein_2=None,
        **kwargs,
    ):
        """Compare two proteins, using protein class as input

        Args:
            protein_1 (Protein, optional): socialgene Protein object. Defaults to None.
            protein_2 (Protein, optional): socialgene Protein object. Defaults to None.
            append (bool, optional): Should results be appended to previously calculated comparisons? Defaults to False.
            verbose (bool, optional): If `True` then additional info will be printed. Defaults to False.
        """

        return self._calculate_mod_score_from_domain_lists(
            protein_id_1=protein_1.uid,
            protein_id_2=protein_2.uid,
            input_list_1=protein_1.domain_vector,
            input_list_2=protein_2.domain_vector,
            **kwargs,
        )

    @staticmethod
    def _calculate_mod_score_from_domain_lists(
        protein_id_1,
        protein_id_2,
        input_list_1,
        input_list_2,
    ):
        return [
            protein_id_1,
            protein_id_2,
            *(mod_score(input_list_1=input_list_1, input_list_2=input_list_2).values()),
        ]

    @staticmethod
    def compare_two_proteins(protein_1, protein_2):
        """
        dict: {l1, l2, levenshtein, jaccard, mod_score}; mod_score -> 2 = Perfectly similar; otherwise (1/Levenshtein + Jaccard)
        Returns:

          protein_2: An instance of a protein object representing the second protein.
          protein_1: An instance of a protein object representing the first protein.
        Args:

        `mod_score` function.
        The function `compare_two_proteins` compares the domain vectors of two proteins using the
        """

        return mod_score(protein_1.domain_vector, protein_2.domain_vector)

    def compare_proteins(
        self,
        uid_list_of_tuples=None,
        cpus=None,
        append=False,
        verbose=False,
    ):
        """Calculate similarity between proteins based on their domain content

        Args:
            uid_list_of_tuples (str, optional): list of two-ples of protein hashes; use `None` to calculate all-vs-all. Defaults to None.
            cpus (int, optional): If >1 then multiprocessing may/may-not be used to speed things up. Defaults to None.
            append (bool, optional): Should results be appended to previously calculated comparisons? Defaults to False.
            verbose (bool, optional): If `True` then additional info will be printed. Defaults to False.
        """
        if uid_list_of_tuples is not None:
            for id_pair in uid_list_of_tuples:
                _temp = self._calculate_mod_score_from_domain_lists(
                    protein_id_1=id_pair[0],
                    protein_id_2=id_pair[1],
                    input_list_1=self.proteins[id_pair[0]].domain_vector,
                    input_list_2=self.proteins[id_pair[1]].domain_vector,
                )
                if append:
                    self.protein_comparison.append(_temp)
                else:
                    return _temp

        elif uid_list_of_tuples is None:
            if len(self.proteins) < 10:
                if verbose:
                    log.info(
                        f"Running all-vs-all comparison of {len(self.proteins)} proteins (list comprehension method)"
                    )
                append_or_not(
                    result_list=self.protein_comparison,
                    result=[
                        mod_score(i1=i1, i2=i2)
                        for i1, i2 in combinations_with_replacement(
                            self.proteins.values(), r=2
                        )
                    ],
                    append=append,
                )
            else:
                if cpus is None:
                    cpus = cpu_count() - 1
                    if cpus < 1:
                        cpus = 1
                if verbose:
                    log.info(
                        f"Running all-vs-all comparison of {len(self.proteins)} proteins (multiprocessing method, using {cpus}-cores)"
                    )
                with Pool(cpus) as p:
                    append_or_not(
                        result_list=self.protein_comparison,
                        result=p.starmap(
                            mod_score,
                            combinations_with_replacement(self.proteins.values(), r=2),
                        ),
                        append=append,
                    )

    def protein_comparison_to_df(self):
        if isinstance(self.protein_comparison, list):
            log.info(
                "Converting protein_comparison list to dataframe and sorting by mod_score"
            )
            self.protein_comparison = pd.DataFrame(
                self.protein_comparison,
                columns=[
                    "query",
                    "target",
                    "query_n_domains",
                    "target_n_domains",
                    "levenshtein",
                    "jaccard",
                    "mod_score",
                ],
            )
            self.protein_comparison.sort_values(
                by=["mod_score"], inplace=True, ascending=False
            )
        elif isinstance(self.protein_comparison, pd.DataFrame):
            log.info(
                "protein_comparison is already a pandas dataframe, skipping conversion"
            )
        else:
            log.info("confused.....protein_comparison isn't list or dataframe")

    def protein_comparison_append_species_and_assemblies(self):
        """Take the pandas DF target column, and search neo4j for those protein ids, return each
        assemblies and species name for each protein. Merge it into the pandas DF.
        """
        result = self.query_neo4j(
            cypher_name="get_species_and_assembly_from_protein_list",
            param=list(set(self.protein_comparison["target"])),
        )
        result = pd.DataFrame(result)

        self.protein_comparison = pd.merge(
            self.protein_comparison,
            result,
            how="left",
            on=None,
            left_on="target",
            right_on="protein",
            left_index=False,
            right_index=False,
            sort=False,
            suffixes=("_x", "_y"),
            copy=True,
            indicator=False,
            validate=None,
        )
        self.protein_comparison.drop(
            "protein",
            axis=1,
            inplace=True,
        )

    def protein_comparison_write_json(self, outpath=None, orient="records"):
        # orient=values is smaller output than orient=records
        if outpath is None:
            return self.protein_comparison.to_json(
                path_or_buf=outpath, orient=orient, double_precision=2, force_ascii=True
            )
        else:
            self.protein_comparison.to_json(
                path_or_buf=outpath, orient=orient, double_precision=2, force_ascii=True
            )

    def protein_comparison_sort(self):
        self.protein_comparison.sort_values(
            by=["mod_score"], inplace=True, ascending=False
        )

    def protein_comparison_append_protein_names(self):
        """Take the pandas DF target column, and search neo4j for those protein ids, return each
        assemblies and species name for each protein. Merge it into the pandas DF.
        """
        result = self.query_neo4j(
            cypher_name="get_protein_name",
            param=list(set(self.protein_comparison["target"])),
        )
        result = pd.DataFrame(result)
        self.protein_comparison = pd.merge(
            self.protein_comparison,
            result,
            how="left",
            on=None,
            left_on="target",
            right_on="external_id",
            left_index=False,
            right_index=False,
            sort=False,
            suffixes=("_x", "_y"),
            copy=True,
            indicator=False,
            validate=None,
        )
        self.protein_comparison.drop(
            "external_id",
            axis=1,
            inplace=True,
        )



import argparse
import csv
import re
from copy import deepcopy
from pathlib import Path

import requests

from socialgene.utils.logging import log

parser = argparse.ArgumentParser(
    description="Download and create edgelist of all Go terms"
)

parser.add_argument(
    "--outdir",
    help="Output directory filepath",
    required=True,
)

OBO_URL = "http://purl.obolibrary.org/obo/go/go-basic.obo"

goterm_pattern = re.compile("GO:[0-9]{7}", flags=0)


def extract_goterm_int_as_str(id):
    if goterm := goterm_pattern.search(id):
        return str(goterm.group().removeprefix("GO:"))


class GoNode:
    __slots__ = ["uid", "name", "namespace", "definition"]

    def __init__(self) -> None:
        self.uid = None
        self.name = None
        self.namespace = None
        self.definition = None

    def add_id(self, x):
        if goterm := extract_goterm_int_as_str(x):
            self.uid = goterm
        else:
            raise ValueError(f"'{x}' isn't a goterm")

    def add_name(self, x):
        self.name = x.removeprefix("name:").strip()

    def add_namespace(self, x):
        self.namespace = x.removeprefix("namespace:").strip()

    def add_definition(self, x):
        self.definition = x.removeprefix("definition:").strip()

    def assign(self, line):
        if line.startswith("id:"):
            self.add_id(line)
        elif line.startswith("name:"):
            self.add_name(line)
        elif line.startswith("namespace:"):
            self.add_namespace(line)
        elif line.startswith("definition:"):
            self.add_definition(line)

    def output(self):
        return (self.uid, self.name, self.namespace, self.definition)


class GoRelationship:
    __slots__ = [
        "start",
        "end",
        "type",
    ]

    def __init__(self) -> None:
        self.start = None
        self.end = None
        self.type = None

    def add_end(self, id):
        if goterm := extract_goterm_int_as_str(id):
            self.end = goterm
        else:
            print(f"'{id}' isn't a goterm")

    def add_type(self, line):
        key = line.split(":", 1)[0]
        if key == "relationship":
            # 'relationship: part_of GO:0000785 ! chromatin'
            # should return: "part of"
            self.type = line.split(":", 1)[1].strip().split(" ", 1)[0]
        elif key == "is_a":
            self.type = key
        self.type = self.type.upper()
        self.end = extract_goterm_int_as_str(line)

    def assign(self, start, line):
        self.start = start
        if line.startswith("is_a") or line.startswith("relationship"):
            self.add_type(line)

    def output(self):
        if all((self.start, self.end, self.type)):
            return (self.start, self.end, self.type)


class AltRel(GoRelationship):
    def __init__(
        self,
    ):
        super().__init__()

    def assign(self, main_id, line):
        if line.startswith("alt_id"):
            self.start = main_id
            self.end = extract_goterm_int_as_str(line)
            self.type = "ALTERNATE"


def write(
    input: list[tuple[str, ...]],
    outpath: str,
) -> None:
    with open(outpath, "w") as h:
        writer = csv.writer(
            h, quoting=csv.QUOTE_MINIMAL, delimiter="\t", dialect="unix"
        )
        writer.writerows(input)


def download_obo(url=OBO_URL):
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        for line in r.iter_lines():
            yield line.decode(r.encoding).strip()


def read_from_file(filepath):
    with open(filepath, "r") as h:
        for i in h:
            yield i.strip()


def parse(outdir, filepath=None, url=None):
    nodes = []
    relationships = []
    if url:
        _open = download_obo(url=url)
    elif filepath:
        _open = read_from_file(filepath=filepath)
    else:
        raise ValueError("Must input url or filepath")
    term_open = False
    for line in _open:
        if line == "[Term]":
            term_open = True
            nodeobj = GoNode()
            relobj = GoRelationship()
        if not term_open:
            continue
        nodeobj.assign(line=line)
        # some goterms have "alt_ids" these are sometimes referenced from other sources but only exist
        # in the obo file as alt ids (no primary entry)
        if line.startswith("alt_id"):
            # deepcopy so name and namespace will carry if present
            alt_node = deepcopy(nodeobj)
            alt_node.add_id(line)
            alt_rel = AltRel()
            alt_rel.assign(main_id=nodeobj.uid, line=line)
            nodes.append(alt_node.output())
            relationships.append(alt_rel.output())
            del alt_node, alt_rel
        relobj.assign(start=nodeobj.uid, line=line)
        if line == "":
            term_open = False
            nodes.append(nodeobj.output())
            # only output full relationhip tuples
            if out := relobj.output():
                relationships.append(out)
    log.info("Writing goterms nodes")
    write(nodes, Path(outdir, "goterms"))
    log.info("Writing goterms relationships")
    write(relationships, Path(outdir, "goterm_edgelist"))
    log.info("Finished reading/writing goterms")


def main():
    args = parser.parse_args()
    parse(url=OBO_URL, outdir=args.outdir)


if __name__ == "__main__":
    main()



from pathlib import Path

import socialgene.utils.file_handling as fh
from socialgene.utils.logging import log


class Domtblout:
    def __init__(self):
        pass

    def parse_domtblout(self, input_path, hmmsearch_or_hmmscan="hmmsearch"):
        """Parse a HMMER domtblout file

        Args:
            input_path (str): path to file
            hmmsearch_or_hmmscan (str): Determines which acc belongs to protein and which belongs to the hmm

        Raises:
            FileNotFoundError: _description_
        """
        # For hmmsearch/hmmscan, HMMER switches the query/target for domtblout, so make that variable here
        for i in self._parse_domtblout(
            input_path=input_path, hmmsearch_or_hmmscan=hmmsearch_or_hmmscan
        ):
            self.add_protein(uid=i["external_id"])
            self.proteins[i["external_id"]].add_domain(**i)

    def _parse_domtblout(self, input_path, hmmsearch_or_hmmscan="hmmsearch"):
        """
        The function `_parse_domtblout` is used to parse a HMMER domtblout file, extracting relevant
        information and returning it as a dictionary.

        Args:
          input_path: The `input_path` parameter is a string that represents the path to the HMMER
        domtblout file that needs to be parsed.
          hmmsearch_or_hmmscan: The parameter `hmmsearch_or_hmmscan` is a string that determines whether
        the input file is from `hmmsearch` or `hmmscan`. Defaults to hmmsearch
        """
        # For hmmsearch/hmmscan, HMMER switches the query/target for domtblout, so make that variable here
        if hmmsearch_or_hmmscan == "hmmscan":
            target_name = "hmm_id"
            query_name = "external_id"
        elif hmmsearch_or_hmmscan == "hmmsearch":
            target_name = "external_id"
            query_name = "hmm_id"
        else:
            raise ValueError
        all_columns = (
            target_name,
            "target_acc",
            "tlen",
            query_name,
            "query_acc",
            "qlen",
            "evalue",
            "seq_pro_score",
            "seq_pro_bias",
            "number",
            "of",
            "c_evalue",
            "i_evalue",
            "domain_score",
            "domain_bias",
            "hmm_from",
            "hmm_to",
            "ali_from",
            "ali_to",
            "env_from",
            "env_to",
            "acc",
            "desc",
        )

        input_path = Path(input_path)
        if not input_path.exists():
            raise FileNotFoundError(input_path)
        with fh.open_read(input_path) as f:
            for line in f:
                if line.startswith("#"):
                    pass
                else:
                    yield dict(zip(all_columns, line.split()))


class ParsedDomtblout:
    """For HMMER domtblout files already processed and rewritten with class Domtblout()"""

    def __init__(self):
        pass

    def parse_parseddomtblout(self, input_path, **kwargs):
        """
        The function `parse_parseddomtblout` reads data from a parsed domtblout TSV and adds protein and domain
        information to a data structure.

        Args:
          input_path: The `input_path` parameter is the path to the input file that contains the data to
        be parsed. It should be a string representing the file path.
        """
        input_path = Path(input_path)
        with fh.open_read(input_path) as f:
            for line in f:
                line = line.strip().split("\t")
                self.add_protein(
                    uid=line[0],
                )
                self.proteins[line[0]].add_domain(
                    uid=line[1],
                    env_from=int(line[2]),
                    env_to=int(line[3]),
                    seq_pro_score=float(line[4]),
                    evalue=float(line[5]),
                    i_evalue=float(line[6]),
                    domain_bias=float(line[7]),
                    domain_score=float(line[8]),
                    seq_pro_bias=float(line[9]),
                    hmm_from=int(line[10]),
                    hmm_to=int(line[11]),
                    ali_from=int(line[12]),
                    ali_to=int(line[13]),
                    exponentialized=bool(line[14]),
                )
        log.info(
            f"Read {str(len(self.proteins))} proteins and {str(sum([len(i.domains) for i in self.proteins.values()]))} domains from {str(input_path)}"
        )


class HmmerParser(Domtblout, ParsedDomtblout):
    def __init__(self):
        super().__init__()

    def parse_hmmout(self, filepath, **kwargs):
        """Parse sequence files (main function)
        Args:
            filepath (str): path to sequence file
        """
        if check_if_domtblout(filepath):
            self.parse_domtblout(filepath, **kwargs)
        elif check_if_parseddomtblout(filepath):
            self.parse_parseddomtblout(**kwargs)
        else:
            raise NotImplementedError(
                "May not be implemented, or you need to use the genbank/fasta parser directly. (e.g. for tar archives)"
            )


def check_if_parseddomtblout(filepath):
    """
    The function `check_if_parseddomtblout` checks if the headers in a file are possibly from a parsed
    domtblout file.

    Args:
      filepath: The `filepath` parameter is a string that represents the path to a SocialGene parsed domtblout file.

    Returns:
      a boolean value, either True or False.
    """
    log.info(filepath)
    with fh.open_read(filepath) as f:
        l1 = f.readline()
        l2 = f.readline()
        a = l1.count("\t")
        b = l2.count("\t")
    return all(i == 14 for i in [a, b])


def check_if_domtblout(filepath):
    """
    The function `check_if_domtblout` checks if the headers of a file match the expected headers of a
    HMMER domtblout file.

    Args:
      filepath: The `filepath` parameter is a string that represents the file path of the file that
    needs to be checked.

    Returns:
      a boolean value (True or False) indicating whether the headers in the file at the given filepath
    match the expected headers for a HMMER domtblout file.
    """
    with fh.open_read(filepath) as f:
        l1 = f.readline()
        l2 = f.readline()
    expected_header = "#---fullsequence-----------------thisdomain-------------hmmcoordalicoordenvcoord\n"
    if l1.replace(" ", "") == expected_header:
        if (
            l2.replace(" ", "")
            == "#targetnameaccessiontlenquerynameaccessionqlenE-valuescorebias#ofc-Evaluei-Evaluescorebiasfromtofromtofromtoaccdescriptionoftarget\n"
        ):
            return True
        else:
            return False
    else:
        return False



import argparse

from rich.progress import Progress, SpinnerColumn, TextColumn

from socialgene.neo4j.schema.graph_schema import GraphSchema
from socialgene.utils.logging import CONSOLE, log

parser = argparse.ArgumentParser(
    description="Add indices to a SocialGene Neo4j Database"
)
parser.add_argument(
    "--all",
    help="",
    default=False,
    required=False,
    action="store_true",
)
parser.add_argument(
    "--labels",
    help="Node labels to try and add indices to",
    default=False,
    required=False,
    nargs="+",
)


def main():
    args = parser.parse_args()
    as_dict = {}
    log.info(
        "Adding indices to Neo4j. This can take some time, depending on how many of each label exists in the database."
    )
    for i in GraphSchema.ALL_NODES:
        if len(i.neo4j_label) == 1:
            as_dict[i.neo4j_label[0]] = i
        else:
            as_dict[i.neo4j_label[-1]] = i
    if not any([i for i in args.__dict__.values()]):
        parser.print_help()
        print(f"Available labels: {sorted(list(as_dict.keys()))}")
        return
    if args.all:
        with Progress(
            SpinnerColumn(spinner_name="runner"),
            TextColumn(text_format="Adding index for... {task.fields[lab]}"),
            console=CONSOLE,
            transient=True,
        ) as progress:
            task = progress.add_task("Adding index for...", lab="")
            for i in GraphSchema.ALL_NODES:
                progress.update(task, lab=i.neo4j_label)
                try:
                    i().add_constraints_to_neo4j()
                    i().add_nonunique_index_to_neo4j()
                except Exception as e:
                    log.warning(f"Failed to add indices to {i.neo4j_label}: {e}")

    elif args.labels:
        with Progress(
            SpinnerColumn(spinner_name="runner"),
            TextColumn(text_format="Adding index for... {task.fields[lab]}"),
            console=CONSOLE,
            transient=True,
        ) as progress:
            task = progress.add_task("Adding index for...", lab="")
            for i in args.labels:
                if i in as_dict:
                    progress.update(task, lab=i)
                    try:
                        as_dict[i]().add_constraints_to_neo4j()
                        as_dict[i]().add_nonunique_index_to_neo4j()
                    except Exception as e:
                        log.warning(f"Failed to add indices to {i}: {e}")
                else:
                    log.warning(f"Label {i} not found in GraphSchema.ALL_NODES")
                    log.warning(f"Available labels: {as_dict.keys()}")


if __name__ == "__main__":
    main()
